package net.pricefx.formulaengine.scripting;

import com.googlecode.genericdao.search.Filter;
import groovy.lang.Closure;
import net.pricefx.common.api.InputType;
import net.pricefx.common.dto.creationworkflow.CreationWorkflowDefinition;
import net.pricefx.common.dto.creationworkflow.CreationWorkflowStepDefinition;
import net.pricefx.domain.*;
import net.pricefx.formulaengine.*;
import net.pricefx.formulaengine.AbstractProducer.ResultIterator;
import net.pricefx.formulaengine.DatamartContext.DataSlice;
import net.pricefx.formulaengine.DatamartRowSet;
import net.pricefx.formulaengine.TableContext;
import net.pricefx.formulaengine.XExpression;
import net.pricefx.formulaengine.scripting.inputbuilder.InputBuilderFactory;
import net.pricefx.server.dto.SummaryCalculableObjectItem;
import net.pricefx.server.dto.calculation.AttributedResult;
import net.pricefx.server.dto.calculation.ConfiguratorEntry;
import net.pricefx.server.dto.calculation.ConfiguratorEntryArray;
import net.pricefx.server.dto.calculation.ContextParameter;
import net.pricefx.server.dto.calculation.DashboardController;
import net.pricefx.server.dto.calculation.ResultFlexChart;
import net.pricefx.server.dto.calculation.ResultGauge;
import net.pricefx.server.dto.calculation.ResultHighchart;
import net.pricefx.server.dto.calculation.ResultHighmap;
import net.pricefx.server.dto.calculation.ResultMatrix;
import net.pricefx.server.dto.calculation.WrappedResult;
import net.pricefx.service.ItemSummaryQuery;
import net.pricefx.workflowengine.WorkflowInfo;
import org.joda.time.DateTime;
import org.joda.time.DateTimeZone;

import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

/**
 * This interface represents what can be reached in the Groovy element syntax as
 * {@code api}. That {@code api} variable or binding is merely an instance of
 * this interface to expose certain "safe" methods and functionality.<br>
 * <br>
 * <b>Caching usage</b><br>
 * The API exposes three ways to store transient and temporary data during formula
 * executions. Each variant has its own specifics and therefore is not suitable for all use cases. The options are:<br>
 * <ul>
 * <li>api.local</li>
 * <li>api.global</li>
 * <li>Shared Cache</li>
 * </ul>
 * <br>
 * The first two can be understood as mere in-memory hash maps. For example,
 * api.local.put("key","value") will work fine.<br>
 * The difference between them is the potential scope and retention of the
 * contained data.<br>
 * <br>
 * <u>api.local</u><br>
 * This is a hash map that is available during the lifetime of exactly one
 * logic execution. It can (and should) be used to store intermediate results
 * that need to be accessed by other subsequent logic elements.<br>
 * <br>
 * <u>api.global</u><br>
 * By default, api.global is very similar to api.local. Without any further
 * settings, the only difference is that api.global is also available in Groovy
 * library functions (while api.local is not).<br>
 * However, by setting api.retainGlobal = true you can instruct the formula engine to keep
 * the hash map in between logic runs. It works between line items and, in case of Quotes,
 * Contracts, Rebate Agreements and Claims, between the header and the line items (but only in the
 * pre-phase, not in the post-phase). The values are also carried forward between two list
 * calculation passes (i.e. form initial run to dirty item run; ONLY in non-distributed mode.).<br>
 * <p><b>Important notes</b>:<br>
 * retainGlobal can be set to true either in the logic or globally in Configuration &gt;
 * All Modules &gt; General Settings &gt; api.retainGlobal defaults to TRUE.<br>
 * retainGlobal is effective only once. In order to keep it for
 * the 3rd item as well (and so on) you need to call it in every formula
 * execution.<br>
 * api.global data also stick to their execution thread. That means that they are NOT shared between
 * the different parallel executions of a distributed calculation. It effectively builds up multiple
 * caches. Another side effect of this is that api.global is not retained and passed to a subsequent
 * calculation pass if the original calculation was in a distributed mode.
 * <br>
 * But even in non-distributed mode api.global is not passed on to the next calculation pass in case the data inside is too large.
 * <br>
 * <br>
 * So generally api.global should be treated as a cache. And the nature of caches is that there is no guarantee entries are there - in particular
 * when expecting this between calculation passes. Formula logic should <b>always</b> handle this situation gracefully when expected cache entries aren't available.
 * <p>
 * Note also that since Vesper 6.0, the maximum string size of JSON serialized api.global content that is
 * passed on to dirty passes is limited to 2 MB. This is configurable in pricefx-config.xml (maxGlobalsSize).
 * <p>
 * <u>Shared Cache</u><br>
 * The main limitation of api.global is the restriction of its availability to
 * the current execution thread. In a distributed calculation, however, there are
 * multiple threads and even multiple servers involved in the calculation. If all
 * these should share a single cache, you must use the shared cache. The shared
 * cache is even available to different calculations (such as other lists or list &amp;
 * CFS or dashboard logic etc). Be careful with key naming so that you do not
 * mix and match data.
 * <p>
 * These advantages come at a cost: Due to the distributed "superglobal" nature,
 * the read/write operations of that cache are slower compared to a real
 * local in-memory map. Use the shared cache only for data that is expensive
 * to calculate or for avoiding frequent database calls. Data in the shared cache
 * expires automatically, the TTL is extended everytime a key is written or read.
 * Values stored are of the String type only. Serialization and deserialization of other
 * objects like dates and numbers are to be done in logic itself.
 * <br><br>
 * <b>Object modifications</b><br>
 * The API contains methods to manipulate data (add, addOrUpdate, delete, etc).
 * These methods can only be used in certain formula contexts (otherwise they are ignored).
 * These contexts are: <br>
 * <ul>
 * <li>From within a CFS execution (non-distributed only)</li>
 * <li>From within a calculation flow</li>
 * <li>From within a direct logic execution via JSON API </li>
 * </ul>
 * <br>
 * On top of this, some restrictions apply which data objects can be manipulated - i.e. not every possible object can be modified.
 * Currently available whitelisted typecodes are:
 * <ul>
 * <li>All customer extension types</li>
 * <li>All product extension types</li>
 * <li>All matrix price parameter values and normal price parameter values (LTV) as well as price parameter header (LT)</li>
 * <li>PGI and XPGI</li>
 * <li>PLI and XPLI</li>
 * <li>MPLI</li>
 * <li>P</li>
 * <li>C</li>
 * <li>PCOMP</li>
 * <li>PDESC</li>
 * <li>PGI</li>
 * <li>TODO</li>
 * <li>Data export, import and archive</li>
 * </ul>
 */
public interface PublicGroovyAPI {

	/**
	 * Returns true if the logic/formula is being executed in the <i>syntax check</i> mode.
	 * The syntax check mode is a special type of execution run used to determine the logic input parameters
	 * in the configuration dialogs, where the logics are selected (e.g. in configuration of LPGs, PLs or dashboards).
	 * These input parameters are also used on document line items (like Quote, Contract, etc).
	 * The syntax check mode runs against no calculable item (in context of no object), so all the functions that depend on item data
	 * (e.g. {@link #product()} or {@link #currentItem()} as well as {@link #getElement(String)}) produce null or mocked values.
	 * <p>
	 * Note: The best practice is to gather all parameters in the first elements of the logic and then abort the calculation
	 * after all parameters are gathered, otherwise all code which depends on the item data will need a special treatment
	 * (it would have to be surrounded using this function).
	 * <p>
	 * Example:
	 * <p>Element Amount:
	 * <blockquote>
	 * <pre>
	 * return api.decimalUserEntry("Amount")
	 * </pre>
	 * </blockquote>
	 * Element Comment:
	 * <blockquote>
	 * <pre>
	 * return api.stringUserEntry("Comment")
	 * </pre>
	 * </blockquote>
	 * Element SyntaxCheckAbort:
	 * <blockquote>
	 * <pre>
	 * if (api.isSyntaxCheck()) {
	 *   api.abortCalculation()
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @return True if the logic is executed in the syntax check mode.
	 */
	boolean isSyntaxCheck();

	/**
	 * Returns true if the logic is being executed via the UI "Test Logic" execution mode.
	 * Because during "testing" the logic does not have all information available, you can use this function to detect
	 * this mode, and mock some necessary data, if they are not available.
	 *
	 * <p>
	 * For example, when you test the Quote line item logic, then you usually need the customerId, which is often
	 * selected only in the header of the Quote. But since you test the Quote line logic standalone, the information
	 * from the header is not available. Using the isDebugMode you can detect this special execution mode and supply/mock
	 * a particular customerId for testing purposes.
	 *
	 * <p>Code sample (Quote line item logic):
	 * <blockquote><pre>
	 * if(api.isDebugMode()) {
	 *     input.Customer = "CD-00004"      // in regular execution input.Customer would be inherited from header input
	 *     input.DeliveryType = "Standard"  // in regular execution input.DeliveryType would be inherited from header input
	 * }
	 *
	 * def country = api.customer("Country", input.Customer)
	 * </pre></blockquote>
	 *
	 *
	 * @return True if the logic is being executed in the "Test Logic" mode.
	 */
	boolean isDebugMode();

	/**
	 * Returns true if the calculation is executed in the distributed mode.
	 * Useful to suppress warnings (or not execute methods) for operations that do not work in the distributed mode.
	 * These operations are basically all write operations that affect other items.
	 * Examples: all object modification ops (add, addOrUpdate, update, delete), marking other items dirty.
	 *
	 * @return True if executed in the distributed mode.
	 */
	boolean isDistributedMode();

	/**
	 * A publicly available cache (as a hashmap) for storing and sharing values across logic elements.
	 * If {@code api.retainGlobal} is set to true, this map will also be available with the entire list calculation,
	 * i.e across item boundaries. It can also be accessed directly as {@code api.global}.
	 *
	 * <p><strong>Note</strong>: This map must be treated as a cache.
	 * It means that the logic should never rely on the cached values, as in various situations this cache is purged
	 * or structured differently. For example, in a distributed calculation, every calculation thread has its own cache;
	 * or in a multi-pass list calculation, the subsequent passes may not have the cache of the previous pass available.
	 * The system always tries to make it available, but the logic should never rely on it.
	 *
	 * @return the hashmap
	 */
	Map<String, Object> getGlobal();

	/**
	 * Puts an entire map of variables into the global cache.
	 *
	 * @param globalVars Map of variables to put into the global cache.
	 */
	void putGlobal(final Map<String, Object> globalVars);

	/**
	 * Retrieves the CalculationContext of the logic evaluation (logic element group).<br><br>
	 * <ol type="a">
	 * <li>For a PA Calculation or Flush logic this can be (from PriceFxInterface):
	 * <ul>
	 *      <li>FORMULAELEMENTGROUP_DM_INIT = "init" - executed before rows</li>
	 *      <li>FORMULAELEMENTGROUP_DM_ROW = "row" - executed for every line of the Target/Source</li>
	 *      <li>FORMULAELEMENTGROUP_DM_SUMMARY = "summary" - executed after rows</li>
     * </ul>
     * </li>
     * <li>In RebateAgreement calculation there is:
	 * <ul>
     *      <li>FORMULAELEMENTGROUP_RM_AGREEMENT = "agreement"</li>
     *      <li>FORMULAELEMENTGROUP_RM_AGREEMENT_READONLY = "agreementReadOnly"</li>
     *      <li>FORMULAELEMENTGROUP_RM_REBATERECORD = "rebateRecord"</li>
     * </ul>
     * </li>
     * <li>In the PriceOptimizer module, when evaluating Model formulas, there is:
     * <ul>
     *      <li>FORMULAELEMENTGROUP_MODEL_DEFINITION = "definition"</li>
     *      <li>FORMULAELEMENTGROUP_MODEL_CALCULATION = "calculation"</li>
     *      <li>FORMULAELEMENTGROUP_MODEL_ANALYSIS = "analysis"</li>
     *      <li>FORMULAELEMENTGROUP_MODEL_FINALIZATION = "finalization"</li>
     *      <li>FORMULAELEMENTGROUP_MODEL_RECORD = "record"</li>
     * </ul>
     * </li>
     * </ol>
     * @return CalculationContext as String.
	 */
    String getCalculationContext();

	/**
	 * Gets the info about the current batch of items being calculated by a calculation thread.
	 * The calculation batch size is defined by a server variable calculationTasks.commitBatchSize and by default it is 200 records.
	 * It is good practice to use the functions from the Shared Library.
	 *
	 * @return List of arrays [sku, key2] for PLI, XPLI, PGI, XPGI contexts or [itemId, typedId] for CFS logics run on P, PX.
	 * @see <a href="https://gitlab.pricefx.eu/accelerators/pricefx-logic/-/blob/6b778b6a682891a330d36c9ae4557be587d632d2/CalculationLogic/SharedLib/elements/BatchUtils.groovy">SharedLib/elements/BatchUtils.groovy</a>
	 */
	public List<String[]> getBatchInfo();

	/**
	 * Returns the ID part of an object. The method tries to be smart and extracts also IDs from maps
	 * or typedId strings.
	 * @param obj Given object (Map).
	 * @return ID of the object.
	 */
	Long getId(Object obj);

	/**
	 * Gets the result value returned by the previous logic element.
	 * <p>
	 * <b>Note</b>: Instead of this function, you can also use {@code out.<ElementName>} (or {@code output.<ElementName>}).
	 * <p><b>Warning</b>: In the syntax check mode, this function returns mock results, not the real ones.
	 *
	 * @param name Name of the logic element whose value you want to retrieve.
	 * @return Value or {@code null} if the element is not found. In the syntax check mode, the result value is mocked.
	 */
	Object getElement(String name);

	/**
	 * The same as {@code api.currentItem("key2")}. It retrieves the secondary key (field "key2") of a matrix item record.<br>
	 * It is applicable in the following contexts:
     * <ul>
     *      <li>Matrix price grid (XPGI.key2)</li>
     *      <li>Matrix price list (XPLI.key2)</li>
     *      <li>Matrix simulation (XSIMI.key2)</li>
     * </ul>
	 *
	 * @return Value of the secondary key or {@code null} if calculating a single item record (= not a matrix item record from the list above).
	 */
	Object getSecondaryKey();

	/**
	 * Gets a named binding variable from a library function.
	 * <p>
	 * Example: In a workflow formula, the workflow can be referenced in a Groovy element directly by the binding/variable "workflow":
	 * {@code workflow.addApprovalStep("First Approval")}.
	 * <p>
	 * However, this does not work from a library function as this script does not have the binding available.
	 * You need to use:
	 * <blockquote>
	 * <pre>
	 * api.getBinding("workflow").addApprovalStep("First Approval")
	 * </pre>
	 * </blockquote>
	 *
	 * @param name Name of the variable
	 * @return Binding
	 */
	Object getBinding(String name);

	/**
	 * Returns an effective/target date of the document being calculated. (It retrieves this
	 * mandatory parameter targetDate from the current execution context.)
	 * <p>The target date is usually {@code new Date()} by default or it is overriden by the user in:
     * <ul>
     *     <li>Quotes - In the Quote view, entered via the Effective Date input parameter.</li>
	 *     <li>Price lists - In the price list creation dialog, entered in the Get Parameters step.</li>
	 *     <li>Price grids - In the price grids overview window, in the Target date column.</li>
	 *     <li>PA calculation Data Loads - In the Data Load definition, at the Calculation tab.</li>
	 *     <li>Calculation flow - In the Calculation flow detail, in the Target date column.</li>
     * </ul>
	 *
	 * <p>Example:
	 * <blockquote><pre>
	 * api.targetDate().format("yyyy-MM-dd")
	 * </pre></blockquote>
	 * @return targetDate date
	 */
	Date targetDate();

	/**
	 * Gets a {@code java.util.Calendar} object initialized with the current targetDate (or a new Date() if no target date is given).
	 * It is used for adding days/months/year and/or setting hours, minutes, seconds.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def months = 6
	 *
	 * def calendar = api.calendar()
	 *
	 * // adds months to the date
	 * calendar.add(Calendar.MONTH, months)
	 * calendar.set(Calendar.HOUR_OF_DAY, startHour)
	 * calendar.set(Calendar.MINUTE, 0)
	 * calendar.set(Calendar.SECOND, 0)
	 *
	 * // return the date 6 months later after targetDate at 00:00:00
	 * calendar.getTime()
	 * </pre>
	 * </blockquote>
	 * @return Initialized Calendar object
	 */
	Calendar calendar();

	/**
	 * Gets a calendar object initialized with {@code date}.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def months = 6
	 * def targetDate = api.targetDate()
	 *
	 * def calendar = api.calendar(targetDate)
	 *
	 * // adds months to the date
	 * calendar.add(Calendar.MONTH, months)
	 * calendar.set(Calendar.HOUR_OF_DAY, startHour)
	 * calendar.set(Calendar.MINUTE, 0)
	 * calendar.set(Calendar.SECOND, 0)
	 *
	 * // returns the date 6 months later after the given date at 00:00:00
	 * calendar.getTime()
	 * </pre>
	 * </blockquote>
	 *
	 * @param date Date to initialize the calendar with.
	 * @return Initialized Calendar object.
	 */
	Calendar calendar(Date date);

	/**
	 * Provides access to the standard {@code java.util.Random} generator instance created for the duration of the calculation task
	 * (= not get re-created for each logic element evaluation).
	 * <p>
	 * Example 1:
	 * <blockquote>
	 * <pre>
	 * def randomPct = (90 + api.random().nextInt(20)) / 100
	 * // randomPct contains random Integer number between 90 (inclusive) and 110 (exclusive)
	 * </pre>
	 * </blockquote>
	 *
	 * <p>
	 * Example 2:
	 * <blockquote>
	 * <pre>
	 * def random = api.random().nextLong()
	 * // random contains random Long number e.g. -2249116201139970752
	 * </pre>
	 * </blockquote>
	 * @return {@code java.util.Random} generator instance created for the duration of the calculation task.
	 * (i.e. does not get re-created for each formula (element) evaluation).
	 * @see <a href="https://docs.oracle.com/javase/8/docs/api/java/util/Random.html">https://docs.oracle.com/javase/8/docs/api/java/util/Random.html</a>
	 */
	Random random();

	/**
	 * @return {@link net.pricefx.domain.DMCalendar DatamartCalendar} object.
	 * @deprecated Use {@link net.pricefx.formulaengine.DatamartContext#calendar} instead.
	 */
	Object datamartCalendar();

	/**
	 * Creates a new DatamartSlice which allows setting filter criteria along the Time, CustomerGroup, ProductGroup
	 * or other dimensions in a Datamart.
	 *
	 * @return Instantiated, empty {@link DataSlice} object.
	 * @deprecated Use {@link DatamartContext#newDatamartSlice()} instead.
	 */
	@Deprecated
	Object newDatamartSlice();

	/**
	 * Creates a new DatamartSlice which allows setting filter criteria along the Time, CustomerGroup, ProductGroup
	 * or other dimensions in a Datamart, initialized with the name of the time dimension field and an optional set of
	 * filter criteria.
	 *
	 * @param dateFieldName Name of the time dimension field.
	 * @param timePeriodsAndProductAndCustomerGroups TimePeriod, CustomerGroup, ProductGroup filters.
	 * @return Initialized {@link DataSlice} object.
	 * @deprecated Use {@link DatamartContext#newDatamartSlice(String, Object...)} instead.
	 */
	Object newDatamartSlice(String dateFieldName, Object... timePeriodsAndProductAndCustomerGroups);

	/**
	 * Translates a CustomerGroup/ProductGroup/DMDataSlice to its equivalent filter representation,
	 * mapping the domain level field names to the corresponding Datamart field names in the process.
	 * <p>
	 * Note: A DMDataSlice object can be instantiated by api.newDatamartSlice, or in a Rebate calculation
	 * context, referenced by the 'calulationBase' binding.
	 * <blockquote>
	 * <pre>
	 * def filter = api.customerGroupEntry("Customer Filter")
	 * return filter != null ? api.datamartFilter(filter) : Filter.custom("1=1")
	 * </pre>
	 * </blockquote>
	 * @param groupOrSlice CustomerGroup/ProductGroup/DMDataSlice to translate into Datamart level filter criteria.
	 * @return Filter object representing either the CustomerGroup/ProductGroup dimension slice or the passed slice in the Datamart.
	 */
	Filter datamartFilter(Object groupOrSlice);

	/**
	 * This method is deprecated and exists only for backward compatibility for older in-production formulas.
	 * Use the Groovy query API instead (see {@link net.pricefx.formulaengine.DatamartContext DatamartContext}).
	 * <p>
	 * Runs a query against a PA Datamart. The query is only executed once in the scope of a calculation job (PL generation etc.).
	 * The query is identified by the first argument. This is also the 'handle' by which datamartLookup refers to the result set of the query.
	 *
	 * @param queryAndDatamartAndProjectionsAndFilters  The first argument is the query ID/handle, subsequent arguments are either DM field names (projections) or filter objects.
	 * @return Numbers of rows in the result set. The actual result set data is stored in memory for a subsequent lookup by datamartLookup(...).
	 * @deprecated Use {@link net.pricefx.formulaengine.DatamartContext#newQuery(Table)}
	 */
	Object datamartQuery(Object... queryAndDatamartAndProjectionsAndFilters);

	/**
	 * This method is deprecated and only exists for backward compatibility for older in-production formulas.
	 * Use the Groovy query API instead (see {@link net.pricefx.formulaengine.DatamartContext DatamartContext}).
	 *<p>
	 * Looks up rows in the result of a datamartQuery(...) call.
	 * @param queryAndProjectionsAndFilters  The first arguments is the ID/handle from a datamartQuery(...) call.  Subsequent arguments refer to projections specified in the datamartQuery call or filter objects leading to additional filtering of rows in the result set.
	 * @return A single primitive value if the result of the lookup results in one row for just one projection.
	 * A list of values if the result is multiple rows for one projection.
	 * A map of (field name, value) pairs for a single row result with multiple projections.
	 * A Matrix2D object for multiple rows and projections.
	 * @deprecated Use {@link net.pricefx.formulaengine.DatamartContext#executeQuery(DatamartContext.Query)},
	 * {@link net.pricefx.formulaengine.DatamartContext#executeSqlQuery(String, Object...)} or
	 * {@link net.pricefx.formulaengine.DatamartContext#streamQuery(DatamartContext.Query)}
	 */
	Object datamartLookup(Object... queryAndProjectionsAndFilters);


	/**
	 * Provides an API for querying and loading PA data.
	 * The most common cases are:
	 * <ul>
	 *     <li>Getting the PA Datamart using {@link DatamartContext#getDatamart(String)} to be further queried.</li>
	 *     <li>Getting the PA Data Source using {@link DatamartContext#getDataSource(String)} to be further queried.</li>
	 *     <li>Getting the PA calendar using {@link DatamartContext#calendar()}.</li>
	 * </ul>
	 *
	 * <p>
	 * Example 1:
	 * <blockquote>
	 * <pre>
	 * def ctx = api.getDatamartContext()
	 *
	 * def dm = ctx.getDatamart("Transactions_DM")
	 *
	 * def query = ctx.newQuery(dm, true)
	 *     .select("ProductGroup")
	 *     .select("SUM(SalesValue)")
	 *     .where(
	 *         Filter.equal("Year", "2019"),
	 *     )
	 *     .groupBy("ProductGroup")
	 *     .orderBy("ProductGroup")
	 *
	 * def result = ctx.executeQuery(query)
	 *
	 * return result?.getData()
	 * </pre>
	 * </blockquote>
	 * <p>
	 * Example 2:
	 * <blockquote>
	 * <pre>
	 * def date = api.parseDate("yyyy-MM-dd", "2018-01-24")
	 * def week = api.getDatamartContext().calendar().getWeek(date)
	 * return week
	 * </pre>
	 * </blockquote>
	 *
	 * @return {@link net.pricefx.formulaengine.DatamartContext DatamartContext}
	 */
	DatamartContext getDatamartContext();

	/**
	 * Provides an SQL query (SELECT only) interface to PA query results.
	 * @return {@link net.pricefx.formulaengine.TableContext TableContext}
	 * @throws Exception if a DB connection cannot be established.
	 */
	TableContext getTableContext() throws Exception;

	/**
	 * A {@link DatamartRowSet} provides access to the rows being loaded or calculated.
	 * <b>Available in a PA DataLoad context only.</b>
	 * @return {@link DatamartRowSet} holding the data rows to load in the target FC.
	 */
	DatamartRowSet getDatamartRowSet();

	/**
	 * A {@link DatamartRowSet} provides access to the rows currently being loaded or calculated. Available in a PA DataLoad context only.<br>
	 * The following Data Load types make use of rowsets:
	 * <ul>
	 * 		<li>Flush - The source rowset gives access to the DataFeed data rows, while the target rowset holds the rows to be loaded in the target DataSource (DS).</li>
	 * 		<li>Calculate - The source rowset gives access to the source rows, which can either come from a DF or from the target FC being calculated.
	 * 						The target rowset holds the rows to be loaded in the target FC. Initially, this rowset will be empty unless the DL.withTargetSnapshot
	 * 						option is set, in which case the rowset will be pre-populated with the target FC rows in the scope of the DL (determined by the DL.filter).</li>
	 * </ul>
	 * <p>
	 * Example 1:
	 * <blockquote>
	 * <pre>
	 * def newRow = [
	 *      "sku" : "1234456",
	 *      "cost" : 123
	 * ]
	 *
	 * def target = api.getDatamartRowSet("target")
	 *
	 * // target is null when running the logic test execute, so you can use trace for debugging purposes
	 * if (target) {
	 *     target.addRow(newRow)
	 * } else {
	 *     api.trace("newRow", null, newRow)
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * <p>
	 * Example 2:
	 * <blockquote>
	 * <pre>
	 * def source = api.getDatamartRowSet("source")
	 * def target = api.getDatamartRowSet("target")
	 *
	 * while (source?.next()) {
	 *     def row = source?.getCurrentRow()
	 *
	 *     // perform the required calculation of row
	 *     target["TotalSales"] = ...
	 *
	 *     // target is null when running the logic test execute, so you can use trace for debugging purposes
	 *     if (target) {
	 *         target.addRow(row)
	 *     }
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @param name Either 'source' to select the source FC or 'target' to select the target FC rowset.
	 * @return {@link DatamartRowSet} holding either the source or target rowset.
	 * @see #getDatamartContext()
	 * @see DatamartContext#newQuery(Table)
	 */
	DatamartRowSet getDatamartRowSet(String name);

	/**
	 * Temporarily switches the current SKU context to another SKU.
	 * This can be used to call functions that rely on a SKU context and do not allow to specify an alternate SKU.
	 * <ul>
	 *     <li>{@link #product()}</li>
	 *     <li>{@link #productCompetition()}</li>
	 *     <li>{@link #productExtension(String)}</li>
	 *     <li>{@link #productXRef()}</li>
	 *     <li>{@link #bom()}</li>
	 *     <li>{@link #bomList()}</li>
	 *     <li>{@link #pricelist(String)}</li>
	 *     <li>{@link #getPricelistItem(String)}</li>
	 * </ul>
	 * The switch is reset by {@link #resetSKUContextSwitch} or automatically at the beginning of every element.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def masterSku = api.productXRef()
	 * def costs
	 *
	 * if (masterSku) {
	 *     api.switchSKUContext(masterSku[0].targetSku)
	 *
	 *     // take the cost from the masterSku item
	 *     costs = api.productExtension("Cost")
	 *
	 *     api.resetSKUContextSwitch()
	 * } else {
	 *     // take the cost from the item being calculated
	 *     costs = api.productExtension("Cost")
	 * }
	 *
	 * if (costs) {
	 *     ....
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @param newSKU New SKU to switch to.
	 * @see #resetSKUContextSwitch()
	 * @see #product(String, String)
	 */
	void switchSKUContext(String newSKU);

	/**
	 * Resets an SKU context switch to the originally specified SKU.
	 * @see #switchSKUContext(String)
	 */
	void resetSKUContextSwitch();

	/**
	 * Retrieves the product object for the SKU currently being calculated (SKU in the context).
	 * The result product object is cached and the next call reads the value from the cached record.
     * <p>Example:
	 * <blockquote>
	 * <pre>
	 * def item = api.product()
	 * def message = checkItem(item)
	 *
	 * def checkItem(item) {
	 *     if (!item.attribute2) {          // Product group
	 *         return "Product group not set"
	 *     } else if (!item.attribute5) {   // Manufacturer
	 *         return "Manufacturer not set"
	 *     }
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @return Full current product Map.
	 * @see #currentItem()
	 * @see #product(String)
	 * @see #product(String, String)
	 * @see #product(String, String, String, Object)
	 */
	Object product();

	/**
	 * Retrieves the given attribute value from the Products table for the SKU being calculated (the SKU in the context).
	 * The product object is cached and the next call reads the value from the cached record.
	 * <p>When reading the predefined columns, use their system names - e.g. label, unitOfMeasure.
	 * <p>When accessing the attribute columns, preferably use their "Name" defined for the column when you set up the table.
	 * You can also use their system names (attribute1, ...), but it does not give you much information about which value you are reading.
	 *
	 * <p><strong>Special use case</strong>: When you call the function with the "sku" attribute, this specific call is optimized
	 * in the backend, so it does not make a lookup in the Products table, but instead returns the SKU value directly from the context.
     *
	 * <p><strong>Side effect</strong>: During the Syntax Check mode, this function triggers creation of the input parameter
	 * 'Product'. But as Product is usually selected in another way (e.g. in a Price List or Quote), this would apply only in rare cases.
	 *
	 * <p>Examples:
	 * <blockquote>
	 * <pre>
	 * def label = api.product("label")
	 * def uom = api.product("unitOfMeasure")
	 * def productGroup = api.product("ProductGroup")  //  api.product("attribute1")
	 * def manufacturer = api.product("Manufacturer") //  api.product("attribute4")
	 * </pre>
	 * </blockquote>
	 * @param attributeName Name of the attribute column whose value you want to retrieve.
	 *                         For the customizable attributeX columns, it can be either in the form
	 *                         "attribute1" or the Name provided in the Rename and Customize Column dialog.
	 * @return Value of the attribute.
	 * @see #currentItem(String)
	 * @see #product()
	 * @see #product(String, String)
	 * @see #product(String, String, String, Object)
	 */
	Object product(String attributeName);

	/**
	 * Retrieves the given attribute value from the Products table for a given SKU.
	 * <p>When reading the predefined columns, use their system names - e.g. label, unitOfMeasure.
	 * <p>When accessing the attribute columns, preferably use their "Name" defined for the column when you set up the table.
	 * You can also use their system names (attribute1, ...), but it does not give you much information about which value you are reading.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def masterSku = ...
	 * def label
	 *
	 * if (masterSku) {
	 *     // take the label from the masterSku item
	 *     label = api.product("label", masterSku)
	 * } else {
	 *    // take label from the item being calculated
	 *    label = api.product("label")
	 * }
	 * </pre>
	 * </blockquote>
	 * @param attributeName Name of the attribute column whose value you want to retrieve.
	 *                         For the customizable attributeX columns, it can be either in the form
	 *                         "attribute1" or the Name provided in the Rename and Customize Column dialog.
	 * @param sku           Product.
	 * @return Value of the attribute (or {@code null} if product not found).
	 * @see #product(String)
	 * @see #product()
	 * @see #product(String, String, String, Object)
	 */
	Object product(String attributeName, String sku);

	/**
	 * Retrieves the given attribute value from the product table for a given SKU.
	 * If the given SKU is null, filterFormulaName and filterFormulaParam are used to apply
	 * the filter logic for the Product picker.
	 * @param attributeName 		Attribute whose value should be returned.
	 * @param sku					Product.
	 * @param filterFormulaName  	Unique name of a formula that will be triggered when the Product picker opens.
	 * @param filterFormulaParam 	Additional data for filterFormula.
	 * @return Value of the attribute (or {@code null} if the product not found).
	 * @see #product(String, String)
	 * @see #product(String)
	 * @see #product()
	 */
	Object product(String attributeName, String sku, String filterFormulaName, Object filterFormulaParam);

	/**
	 * Triggers a custom-named product picker input parameter.
	 *
	 * @param parameterName Name of the parameter.
	 * @param attributeName Attribute whose value should be returned ({@code null} for the entire object).
	 * @return Attribute value
	 */
	Object otherProduct(String parameterName, String attributeName);

	/**
	 * Triggers a custom-named product picker input parameter.
	 *
	 * @param parameterName Name of the parameter.
	 * @param attributeName Attribute whose value should be returned ({@code null} for the entire object).
	 * @param filterFormulaName  	Unique name of a formula that will be triggered when the Product picker opens.
	 * @param filterFormulaParam 	Additional data for filterFormula.
	 * @return Attribute value
	 */
	Object otherProduct(String parameterName, String attributeName, String filterFormulaName, Object filterFormulaParam);

	/**
	 * Triggers a custom-named product picker input parameter.
	 *
	 * @param parameterName Name of the parameter.
	 * @return Product object
	 */
	Object otherProduct(String parameterName);

	/**
	 * Retrieves a list of rows from a Product Extension (PX) table named {@code extensionName} for the SKU being
	 * calculated (SKU in the context).
	 * <p>
	 * For performance reasons, consider use of {@link #getBatchInfo()}.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def promotions = api.productExtension("Promotion")
	 * def promotionSize = promotions?.getAt(0)
	 * </pre>
	 * </blockquote>
	 * @param extensionName Name of the Product Extension table to read data from.
	 * @return List of PX rows/records (represented as a Map), so it returns a List of Maps.
	 * @see #getBatchInfo()
	 * @see #productExtension(String, Filter...)
	 */
	Object productExtension(String extensionName);

	/**
	 * Retrieves a list of rows from a Product Extension (PX) table named {@code extensionName} for the SKU being
	 * calculated (SKU in the context) additionally filtered by {@code filters}.
	 * <p>
	 * For performance reasons, consider use of {@link #getBatchInfo()}.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def filter = [
	 *         Filter.greaterOrEqual("ValidFrom", targetDate),   // or Filter.greaterOrEqual("attribute1", targetDate)
	 *         Filter.lessOrEqual("ValidTo", targetDate)       // Filter.lessOrEqual("attribute2", targetDate)
	 * ]
	 *
	 * def promotions = api.productExtension("Promotion", *filter)
	 * def promotion = promotions?.getAt(0)
	 * </pre>
	 * </blockquote>
	 *
	 * @param extensionName Name of the product extension table.
	 * @param filters       One or more {@link Filter} objects that narrow down the search.
	 * @return List of matching PX rows/records (represented as a Map), so it returns a List of Maps.
	 * @see #getBatchInfo()
	 * @see #productExtension(String)
	 */
	Object productExtension(String extensionName, Filter... filters);

	/**
	 * Retrieves a list of all product competition rows/records (PCOMP) for the SKU being calculated (SKU in the context).
	 * <p>
	 * Also, consider use of {@link #productExtension(String)} instead since that gives you more flexibility.
	 *
	 * @return List of PCOMP records (represented as a Map).
	 * @see #getBatchInfo()
	 */
	Object productCompetition();

	/**
	 * Retrieves a list of product competition rows/records (PCOMP) for the SKU being calculated (SKU in the context) filtered by {@code filters}.
	 * Consider use of {@link #productExtension(String)} instead since that gives you more flexibility.
	 *
	 * @param filters One or more {@link Filter} objects that narrow down the search.
	 * @return List of matching PX records (represented as a Map).
	 */
	Object productCompetition(Filter... filters);

	/**
	 * Retrieves a list of all product reference records (PXREF) for the SKU being calculated.
	 * For performance reasons consider use of {@link #getBatchInfo()}.
	 *
	 * @return List of PXREF records (represented as a Map).
	 * @see #getBatchInfo()
	 */
	Object productXRef();

	/**
	 * Prompts the user to enter parameters for a rebate agreement (RBA).
	 *
	 * @param parameterName - Name of an entry in the input parameters listing (mandatory).
	 * @param attributeName - Name of the attribute you need (mandatory).
     * @return Attribute of the rebate agreement or {@code null} if RA does not exist.
     */
	Object rebateAgreementUserEntry(String parameterName, String attributeName);

	/**
	 * Retrieves a list of all product reference records (PXREF) for the SKU being calculated.
	 *
	 * @param filters One or more {@link Filter} objects that narrow down the PXREF search.
	 * @return List of matching PXREF records (represented as a Map).
	 * @see #getBatchInfo()
	 */
	Object productXRef(Filter... filters);

	/**
	 * Retrieves the given attribute value from the Customers table for the customer selected by the user in the input parameter "Customer".
	 * The customer object is cached and the next call reads the value from the cached record.
	 * <p>
	 * <b>Note</b>: The value of the Customer input parameter can be also inherited from the parent; for example, if the user
	 * selects the Customer in the Quote header, then such input value is available also during the Quote Line calculation.
	 * <p>
	 * <b>Important side effect</b>: During the Syntax Check mode, this function triggers creation of the input parameter
	 * Customer. This impacts, for example, the Quote line item where you might need to read a certain customer
	 * attribute value, but maybe not create the Customer picker. In that case you must avoid running this function call
	 * during the Syntax Check mode.
	 *
	 * @param attributeName Attribute value to be returned. Can be either the real column name (like customerId, name, attribute1) or also the name given in the Rename and Customize column dialog.
	 * @return Value of the attribute
	 */
	Object customer(String attributeName);

	/**
	 * Equivalent of the standard syntax "Customer" function. It retrieves the given attribute value of the customer context parameter.
	 * If the value is {@code null} and defaultFromParent is {@code true}, it retrieves (and returns) the value from
	 * the customer's parent regardless if that value is {@code null} as well or not. If there is no parent customer set,
	 * {@code null} will be returned.
	 *
	 * @param attributeName     Attribute value to be returned.
	 * @param defaultFromParent Determines whether to look up the value from the parent customer (if any) in case it is {@code null} in the customer.
	 * @return Value of the attribute
	 */
	Object customer(String attributeName, boolean defaultFromParent);

	/**
	 * Retrieves the given attribute value from the Customers table for the given customer.
	 * The customer object is cached and the next call reads the value from the cached record.
	 *
	 * @param attributeName Attribute value to be returned. Can be either the real column name (like customerId, name, attribute1) or also the name given in the Rename and Customize column dialog.
	 * @param customerId    customerId value to be looked up in the Customers table.
	 * @return Value of the attribute (or {@code null} if no customer found).
	 */
	Object customer(String attributeName, String customerId);

	/**
	 * Retrieves the given attribute value of the given customer. If the value is
	 * {@code null} and defaultFromParent is {@code true}, it retrieves (and returns) the value from the customer's
	 * parent regardless if that value is {@code null} as well or not. If there is no parent customer set, {@code null} will be returned.
	 *
	 * @param attributeName     Attribute value to be returned.
	 * @param customerId        customerId value for the customer search.
	 * @param defaultFromParent Determines whether to look up the value from the parent customer (if any) in case it is {@code null} in the customer.
	 * @return Value of the attribute (or {@code null} if no customer found).
	 */
	Object customer(String attributeName, String customerId, boolean defaultFromParent);

	/**
	 * Retrieves a named attribute value of the named customer (if the value is
	 * {@code null} and defaultFromParent is {@code true}, retrieves (and returns) the value from the customer's
	 * parent regardless if that value is {@code null} as well or not. If there is no parent customer set, {@code null} will be returned)
	 *
	 * @param attributeName      Attribute value to be returned.
	 * @param customerId         customerId value for the customer search.
	 * @param defaultFromParent  Determines whether to look up the value from the parent customer (if any) in case it is {@code null} in the customer.
	 * @param filterFormulaName  Unique name of a formula that will be triggered when the Customer picker opens.
	 * @param filterFormulaParam Additional data for filterFormula (e.g. agreement's typedId,...).
	 * @return Value of the attribute (or {@code null} if no customer found).
	 */
	Object customer(String attributeName, String customerId, boolean defaultFromParent, String filterFormulaName, Object filterFormulaParam);

	/**
	 * Triggers a custom-named customer picker input parameter.
	 *
	 * @param parameterName Name of the parameter.
	 * @return The customer object
	 */
	Object otherCustomer(String parameterName);

	/**
	 * Triggers a custom-named customer picker input parameter.
	 *
	 * @param parameterName Name of the parameter.
	 * @param attributeName Attribute whose value to return ({@code null} for the entire object).
	 * @return Attribute value
	 */
	Object otherCustomer(String parameterName, String attributeName);

	/**
	 * Triggers a custom-named customer picker input parameter.
	 *
	 * @param parameterName     Name of the parameter.
	 * @param attributeName     Attribute whose value should be returned ({@code null} for the entire object).
	 * @param defaultFromParent Determines whether to look up the attribute value from the parent customer (if any) in case it is {@code null} in the customer.
	 * @return Attribute value
	 */
	Object otherCustomer(String parameterName, String attributeName, boolean defaultFromParent);


	/**
	 * Triggers a custom-named customer picker input parameter.
	 *
	 * @param parameterName     Name of the parameter.
	 * @param attributeName     Attribute whose value should be returned ({@code null} for the entire object).
	 * @param defaultFromParent Determines whether to look up the attribute value from the parent customer (if any) in case it is {@code null} in the customer.
	 * @param filterFormulaName  Unique name of a formula that will be triggered when the Customer picker opens.
	 * @param filterFormulaParam Additional data for filterFormula (e.g. agreement's typedId,...).
	 * @return Attribute value
	 */
	Object otherCustomer(String parameterName, String attributeName, boolean defaultFromParent, String filterFormulaName, Object filterFormulaParam);

	/**
	 * Retrieves a list of all customer extentions records (CX) from a CX table named {@code extensionName}.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def filter = Filter.equal("attribute1", "ShipTo")   // Type
	 *
	 * def addresses = api.customerExtension("Addresses", filter)
	 *
	 * if (addresses) {
	 *     def address = addresses.getAt(0)
	 *     ...
	 * }
	 * </pre>
	 * </blockquote>
	 * @param extensionName Name of the extension.
	 * @param filters       None, one or more {@link Filter} objects that narrow down the CX search.
	 * @return List of matching CX records (represented as a Map).
	 */
	Object customerExtension(String extensionName, Filter... filters);

	/**
	 * Returns a line item of the given calculable line item collection (CLIC).
	 * Currently, there are three types of CLICs: Quote, Contract and Rebate Agreement.
	 *
	 * @param clic   CLIC to which the line item belongs, it may either be the CLIC object as such (its map representation to be exact) or typedId or a uniqueName.
	 * @param lineId lineId of the line item.
	 * @param <T> ContractLineItem (CTLI), QuoteLineItem (QLI) or RebateAgreementLineItem (RBALI).
	 * @return Calculable line item
	 */
	Object getCalculableLineItem(Object clic, String lineId);

	/**
	 * Returns results of the given names (or labels) from a line item object (its map representation).
	 * <p>
	 * Note: The line item object will typically be fetched via {@link #getCalculableLineItem(Object, String)}.
	 *
	 * @param cli Calculable line item.
	 * @param resultNamesOrLabels Result names or labels.
	 * @return Either a single result or a map(nameOrLabel,result) of results if there were more names (labels) specified.
	 */
	Object getCalculableLineItemResult(Object cli, String... resultNamesOrLabels);

	/**
	 * Returns the full object (header and line items) of e.g. a quote, rebate agreement or contract.
	 *
	 * @param typedId typeId of the object.
	 * @return Map specifying the line item collection or {@code null} if an invalid typedId was passed.
	 */
	Object getCalculableLineItemCollection(String typedId);

	/**
	 * Creates a revision of CalculableLineItemCollection (Quote, Rebate Agreement, Contract).
	 *
	 * @param typedId typeId of the object.
	 * @return Map specifying the revised line item collection or {@code null} if an invalid typedId was passed.
	 */
	Object createNewRevision(String typedId);

	/**
	 * Adds a warning to the logic/formula execution warning log.
	 * Warnings usually explain to the user that there was a "technical" problem (e.g. missing data, inputs) and the calculation could not be completed.
	 *
	 * <p>
	 * The warnings are displayed for:
	 * <ul>
	 *     <li>Quotes, Rebates, Contracts in the Calculation Results section for the line items. The warning is shown
	 *     next to the element in which it was added, but if the element is not visible, then it will appear by
	 *     the next visible element.</li>
	 *     <li>Price lists, price grids, Datamarts, policy records - all warnings from all elements are listed together
	 *     in a special Warnings column in the grid list items.</li>
	 * </ul>
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def cost = ...
	 *
	 * if (cost == null) {
	 *      api.addWarning("Cannot calculate List Price because of missing cost")
	 *      return null
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @param message Warning message.
	 */
	void addWarning(String message);

	/**
	 * Retrieves the previous calculation result by the logic element label (not the element result name).
	 * @param label Logic element label.
	 * @return Calculation result of that element
	 */
	Object contextByLabel(String label);

	/**
	 * Searches for a record in the price parameter table named {@code parameterName}
	 * where the column 'name' matches the value entered by the user (in a drop-down input parameter) and returns
	 * the value from the column 'value'.
	 * <p><b>Side effect</b>: During the Syntax Check mode this function creates a new input parameter (drop-down)
	 * whose values are taken from the column "Name" of the given Price Parameter.
	 * @param parameterName Name of the price parameter table to look up.
	 * @return For a simple price parameter, it returns the value of the "Value" column from the found row. In case of matrix lookups, the entire object/row.
	 * @see #findLookupTableValues(String, Filter...)
	 */
	Object vLookup(String parameterName);

	/**
	 * Searches for a record in the price parameter table named {@code parameterName}
	 * where the column 'name' (for LTV or MLTV) or 'key1' (for MLTV2-6)
	 * matches the attributeNameOrKeyOrRangeValue value
	 * and returns the value from the column 'value' (LTV) or 'attribute1' (MLTV).
	 * Once the record is found, it remains cached during the whole calculation and
	 * the next call for the same arguments reads the value from the cache.
	 *
     * <p>
     * Example:
	 * <blockquote>
	 * <pre>
	 * def productGroup = api.product("ProductGroup")  // or "attribute2"
	 * def shippingCost = api.vLookup("ShippingCosts", productGroup)
	 * </pre>
	 * </blockquote>
     *
	 * @param parameterName                  Name of the price parameter table.
	 * @param attributeNameOrKeyOrRangeValue Predefined key or range value or in case of a matrix lookup a user entry.
	 * @return Value of the column 'value' for SIMPLE (LTV) or 'attribute1' for MATRIX (MLTV).
	 * @see #findLookupTableValues(String, Filter...)
	 */
	Object vLookup(String parameterName, Object attributeNameOrKeyOrRangeValue);

	/**
	 * Only relevant for MATRIX 1-key price parameter tables (MLTV).
	 * <p>
	 * Searches for a record in the price parameter table named {@code parameterName}
	 * where the column 'key1' matches the {@code key1} parameter
	 * and returns the value of the column {@code attributeName}.
	 * Once the record is found, it remains cached during the whole calculation and
	 * the next call for the same arguments reads the value from the cache.
	 *
	 * <p>
     * Example:
	 * <blockquote>
	 * <pre>
	 * def prevailing = api.vLookup("Manufacturer", "attribute1", "Meat company")
	 * </pre>
	 * </blockquote>
     *
	 * @param parameterName Name of the price parameter table.
	 * @param attributeName Name of the attribute to retrieve or a comma separated list of attributes.
	 * @param key           Key of the parameter table.
	 * @return Value of the column {@code attributeName} or a full object if the attribute name is null.
	 * @see #findLookupTableValues(String, Filter...)
	 */
	Object vLookup(String parameterName, String attributeName, String key);

	/**
	 * Only relevant for MATRIX 2-key price parameter tables (MLTV2).
	 * <p>
	 * Searches for a record in a price parameter table named {@code parameterName}
	 * where the columns 'key1'-'key2' match the {@code key1}-{@code key2} parameters
	 * and returns the value from the column {@code attributeName}.
	 * Once the record is found, it remains cached during the whole calculation and
	 * the next call for the same arguments reads the value from the cache.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def exchangeRate = api.vLookup("ExchangeRate", "attribute1", "EUR", "USD")
	 * </pre>
	 * </blockquote>
	 *
	 * @param parameterName Name of the price parameter table.
	 * @param attributeName Name of the attribute to retrieve or a comma separated list of attributes.
	 * @param key1          key1 of the parameter table.
	 * @param key2          key2 of the parameter table.
	 * @return Value of the column {@code attributeName} or a full object if the attribute name is null.
	 * @see #findLookupTableValues(String, Filter...)
	 */
	Object vLookup(String parameterName, String attributeName, String key1, String key2);

	/**
	 * Only relevant for MATRIX 3-key price parameter tables (MLTV3).
	 * <p>
	 * Searches for a record in the price parameter table named {@code parameterName}
	 * where the columns 'key1'-'key3' match the {@code key1}-{@code key3} parameters
	 * and returns the value from the column {@code attributeName}.
	 * Once the record is found, it remains cached during the whole calculation and
	 * the next call for the same arguments reads the value from the cache.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def exchangeRate = api.vLookup("ExchangeRate", "attribute1", "EUR", "USD", "2018")
	 * </pre>
	 * </blockquote>
	 *
	 * @param parameterName Name of the price parameter table.
	 * @param attributeName Name of the attribute to retrieve or a comma separated list of attributes.
	 * @param key1          key1 of the parameter table.
	 * @param key2          key2 of the parameter table.
	 * @param key3          key3 of the parameter table.
	 * @return Value of the column {@code attributeName} or a full object if the attribute name is null.
	 * @see #findLookupTableValues(String, Filter...)
	 */
	Object vLookup(String parameterName, String attributeName, String key1, String key2, String key3);

	/**
	 * Only relevant for MATRIX 4-key price parameter tables (MLTV4).
	 * <p>
	 * Searches for a record in the price parameter table named {@code parameterName}
	 * where the columns 'key1'-'key4' match the {@code key1}-{@code key4} parameters
	 * and returns the value from the column {@code attributeName}.
	 * Once the record is found, it remains cached during the whole calculation and
	 * the next call for the same arguments reads the value from the cache.
	 *
	 * @param parameterName Name of the price parameter table.
	 * @param attributeName Name of the attribute to retrieve or a comma separated list of attributes.
	 * @param key1          key1 of the parameter table.
	 * @param key2          key2 of the parameter table.
	 * @param key3          key3 of the parameter table.
	 * @param key4          key4 of the parameter table.
	 * @return Value of the column {@code attributeName} or a full object if the attribute name is null.
	 * @see #findLookupTableValues(String, Filter...)
	 */
	Object vLookup(String parameterName, String attributeName, String key1, String key2, String key3, String key4);

	/**
	 * Only relevant for MATRIX 5-key price parameter tables (MLTV5).
	 * <p>
	 * Searches for a record in the price parameter table named {@code parameterName}
	 * where the columns 'key1'-'key5' match the {@code key1}-{@code key5} parameters
	 * and returns the value from the column {@code attributeName}.
	 * Once the record is found, it remains cached during the whole calculation and
	 * the next call for the same arguments reads the value from the cache.
	 *
	 * @param parameterName Name of the price parameter table.
	 * @param attributeName Name of the attribute to retrieve  or a comma separated list of attributes.
	 * @param key1          key1 of the parameter table.
	 * @param key2          key2 of the parameter table.
	 * @param key3          key3 of the parameter table.
	 * @param key4          key4 of the parameter table.
	 * @param key5          key5 of the parameter table.
	 * @return Value of the column {@code attributeName} or a full object if the attribute name is null.
	 * @see #findLookupTableValues(String, Filter...)
	 */
	Object vLookup(String parameterName, String attributeName, String key1, String key2, String key3, String key4, String key5);

	/**
	 * Only relevant for MATRIX 6-key price parameter tables (MLTV6).
	 * <p>
	 * Searches for a record in the price parameter table named {@code parameterName}
	 * where the columns 'key1'-'key6' match the {@code key1}-{@code key6} parameters
	 * and returns the value from the column {@code attributeName}.
	 * Once the record is found, it remains cached during the whole calculation and
	 * the next call for the same arguments reads the value from the cache.
	 *
	 * @param parameterName Name of the price parameter table.
	 * @param attributeName Name of the attribute to retrieve or a comma separated list of attributes.
	 * @param key1          key1 of the parameter table.
	 * @param key2          key2 of the parameter table.
	 * @param key3          key3 of the parameter table.
	 * @param key4          key4 of the parameter table.
	 * @param key5          key5 of the parameter table.
	 * @param key6          key6 of the parameter table.
	 * @return Value of the column {@code attributeName} or a full object if the attribute name is null.
	 * @see #findLookupTableValues(String, Filter...)
	 */
	Object vLookup(String parameterName, String attributeName, String key1, String key2, String key3, String key4, String key5, String key6);

    /**
     * This syntax can be applied to all MATRIX lookup types.
     * The keys are passed in a map.
     * The key name can be the technical name ("key1" etc.) or the meta-renamed name.
	 * <p>Example:
	 * <blockquote><pre>
	 * def keys = [
	 *         "Country": country,
	 *         "DeliveryType": input.DeliveryType
	 * ]
	 *
	 * return api.vLookup("FreightSurcharge", "FreightSurcharge", keys)
	 * </pre></blockquote>
     *
     * @param parameterName Name of the price parameter table.
     * @param attributeName Name of the attribute to retrieve or a comma separated list of attributes.
     * @param keys Map with the keys of the row to retrieve. Note: All keys need to be present for the table in question.
     * @return Map with values of the columns of {@code attributeNames} as keys or a full object if attributeNames is null.
     */
    Object vLookup(String parameterName, String attributeName, Map keys);

	/**
	 * This syntax can be applied to all MATRIX lookup types.
	 * The keys are passed in a map.
	 * The key name can be the technical name ("key1" etc.) or the meta-renamed name.
	 *
	 * @param parameterName Name of the price parameter table.
	 * @param attributeNames Names of the attribute to retrieve.
	 * @param keys Map with the keys of the row to retrieve. Note: All keys need to be present for the table in question.
	 * @return Map with values of the columns of {@code attributeNames} as keys or a full object if attributeNames is null.
	 */
	Map<String,Object> vLookup(String parameterName, List<String> attributeNames, Map keys);

	/**
	 * Returns a numeric value entered by the user in the input parameter.
	 * <p><b>Side effect</b>: During the Syntax Check mode, this function triggers creation of the input parameter.
	 *
     * <ul>
     * <li>When the user-entered number is a whole number (integer),
     * the value returned is java.lang.Integer.</li>
     * <li>When the user-entered number is a decimal number,
     * the value returned is java.math.BigDecimal.</li>
     * <li>If you always need the result as a BigDecimal, you can cast it,
     * e.g. api.decimalUserEntry("number") as BigDecimal.</li>
     * </ul>
	 *
	 * @param entryName Name of the user input parameter. For the names use the "camel case" (no spaces and capital letters).
	 *                     This value is not meant as a label of the input field, but rather a name which should stay
	 *                     stable for a long time. Remember, the input fields are stored e.g. on every single
	 *                     Quote line item and are created only once when adding the new line. If you need to modify
	 *                     the label (or other properties) of the input parameter, you can use {@link #getParameter(String)}.
	 *
	 * @return Numeric value entered by the user. Either of the type Integer or BigDecimal, depending on the value entered.
	 * 	       During SyntaxCheck execution mode, it returns a mock value.
	 * @see #input(String)
	 * @see #stringUserEntry(String)
	 * @see #textUserEntry(String) (String...)
	 * @see #decimalUserEntry(String)
	 * @see #dateUserEntry(String)
	 * @see #dateRangeUserEntry(String)
	 * @see #dateTimeUserEntry(String)
	 * @see #booleanUserEntry(String)
	 * @see #option(String, Object...)
	 * @see #options(String, Object...)
	 * @see #filterBuilderUserEntry(String, String)
	 * @see #datamartFilterBuilderUserEntry(String, String, Object...)
	 * @see #otherProduct(String)
	 * @see #otherCustomer(String)
	 * @see #productGroupEntry(String...)
	 * @see #customerGroupEntry(String...)
	 * @see #rebateAgreementUserEntry(String, String)
	 * @see #anyUser(String)
	 * @see #inputMatrix(String, String...)
	 */
	Object userEntry(String entryName);

	/**
	 * Creates an input parameter that lets the user enter a decimal value.
	 * Rendered as a numeric field.
	 *
	 * @param entryName Name of the user entry parameter.
	 * @return Decimal value entered by the user as a BigDecimal
	 * @see #integerUserEntry(String)
	 */
	Object decimalUserEntry(String entryName);

	/**
	 * Returns an integer numeric value entered by the user in the input parameter.
	 * <p><b>Side effect</b>: During the Syntax Check mode, this function triggers creation of the input parameter
	 * that lets the user enter an integer value. The input is rendered as a numeric field.
	 *
	 * @param entryName Name of the user input parameter. Use "camel case" for the names (no spaces and capital letters).
	 *                     This value is not meant as a label of the input field, but rather a name which should stay
	 *                     stable for a long time. Remember, the input fields are stored e.g. on every single
	 *                     Quote line item and are created only once, when adding the new line. If you need to modify
	 *                     the label (or other properties) of the input parameter, you can use {@link #getParameter(String)}.
	 *
	 * @return Integer value entered by the user as a BigDecimal.
	 * @see #decimalUserEntry(String)
	 * @see #userEntry(String)
	 */
	Object integerUserEntry(String entryName);

	/**
	 * Returns a text value entered by the user in the input parameter.
	 * <p><b>Side effect</b>: During the Syntax Check mode, this function triggers creation of the single-text input parameter.
	 *
	 * @param entryName Name of the user entry parameter.
	 * @return Text value entered by the user as a String.
	 * @see #textUserEntry(String)
	 */
	Object stringUserEntry(String entryName);

	/**
	 * Creates an input parameter that lets the user enter a short text value.
	 * Rendered as a single-line text field.
	 *
	 * @param entryName Name of the user entry parameter.
	 * @param textMask  Optional text mask that is displayed in the UI.
	 * @return Text value entered by the user as a String.
	 * @see #textUserEntry(String)
	 */
	Object stringUserEntry(String entryName, String textMask);

	/**
	 * Creates an input parameter that lets the user enter a short text value.
	 * Rendered as single-line text field.
	 *
	 * @param entryName Name of the user entry parameter.
	 * @param textMask  Optional text mask that is displayed in the UI.
	 * @param label		Label of the entry box.
	 * @return Text value entered by the user as String.
	 * @see #textUserEntry(String, String)
	 */
	Object stringUserEntry(String entryName, String textMask, String label);

	/**
	 * Creates an input parameter that lets the user enter a longer text value.
	 * Rendered as a multi-line text box.
	 *
	 * @param entryName Name of the user entry parameter.
	 * @return Text value entered by the user as a String.
	 * @see #stringUserEntry(String)
	 */
	Object textUserEntry(String entryName);

	/**
	 * Creates an input parameter that lets the user enter a longer text value.
	 * Rendered as a multi-line text box.
	 *
	 * @param entryName Name of the user entry parameter.
	 * @param label		Label of the entry box.
	 * @return Text value entered by the user as a String.
	 * @see #stringUserEntry(String, String, String)
	 */
	Object textUserEntry(String entryName, String label);

	/**
	 * Creates an input parameter that lets the user enter a boolean (yes or no) value.
	 * Rendered as a checkbox.
	 *
	 * @param entryName Name of the user entry parameter.
	 * @return Boolean value entered by the user.
	 */
	Object booleanUserEntry(String entryName);

	/**
	 * Creates an input parameter that lets the user enter a date value.
	 * Rendered as a date picker field.
	 *
	 * @param entryName Name of the user entry parameter.
	 * @return String value entered by the user in the format yyyy-MM-dd.
	 * @see #parseDate(String, String)
	 * @see #timeUserEntry(String)
	 * @see #dateTimeUserEntry(String)
	 * @see #dateRangeUserEntry(String)
	 */
	Object dateUserEntry(String entryName);

	/**
	 * Creates an input parameter that lets the user enter a date range (start and end date).
	 * Rendered as two date picker fields.
	 * <p>
	 * Note: This is available in Unity UI only. No backporting planned for the Classic UI.
	 *
	 * @param entryName Name of the user entry parameter.
	 * @return String array or list with values entered by the user in the format yyyy-MM-dd.
	 * @see #parseDate(String, String)
	 * @see #timeUserEntry(String)
	 * @see #dateTimeUserEntry(String)
	 * @see #dateUserEntry(String)
	 */
	Object dateRangeUserEntry(String entryName);
	/**
	 * Creates an input parameter that lets the user to enter a time value in format "HH:mm" (e.g. "23:59").
	 * Rendered as a time picker field.
	 *
	 * @see #parseDate(String, String)
	 * @param entryName Name of the user entry parameter.
	 * @return String value entered by the user in the format HH:mm.
	 * @see #dateUserEntry(String)
	 * @see #dateTimeUserEntry(String)
	 * @see #dateRangeUserEntry(String)
	 */
	Object timeUserEntry(String entryName);

	/**
	 * Creates an input parameter that lets the user to enter a date time value in format "dd/MM/yyyy HH:mm" (e.g. "23/01/2019 10:30").
	 * Rendered as a datetime picker field.
	 * <p>
	 * Note: The format implicitly uses the GMT/UTC time zone and so the returned value may
	 * differ from what was entered on the client side (as the client time may be in a different time zone).
	 *
	 * @see #parseDate(String, String)
	 * @param entryName Name of the user entry parameter.
	 * @return String value entered by the user in the format "yyyy-MM-dd'T'HH:mm:ss".
	 */
	Object dateTimeUserEntry(String entryName);

	/**
	 * Creates an input parameter that lets the user select a group of products and returns the selection as a Map.
	 * This is commonly used in PromotionManager and RebateManager.
	 * It is rendered as a product group picker widget.
	 *
	 * @param entryName Name of the user entry parameter.
	 * @return Entered product group as a Map.
	 * <pre>
	 * [
	 *   "productFieldName"  : "attribute3",
	 *   "productFieldLabel" : "Business Unit",
	 *   "productFieldValue" : "MeatBall"
	 * ]
	 * </pre>
	 */
	Object productGroupEntry(String... entryName);

	/**
	 * Creates an input parameter that lets the user select a group of products and returns the selection as a Map.
	 * This is commonly used in PromotionManager and RebateManager.
	 * It is rendered as a product group picker widget.
	 * <p>
	 * Only PA searchable Product Extensions will be available.
	 *
	 * @param entryName Name of the user entry parameter.
	 * @return Entered product group as a Map.
	 * <pre>
	 * [
	 *   "productFieldName"  : "attribute3",
	 *   "productFieldLabel" : "Business Unit",
	 *   "productFieldValue" : "MeatBall"
	 * ]
	 * </pre>
	 */
	Object datamartProductGroupEntry(String... entryName);

	/**
	 * Creates an input parameter that lets the user select a group of products and returns the selection as a Map.
	 * This is commonly used in PromotionManager and RebateManager.
	 * It is rendered as a product group picker widget.
	 *
	 * @param entryName Name of the user entry parameter.
	 * @param filterFormulaName Unique name of a formula that will be triggered when the customer picker opens.
	 * @param filterFormulaParam Additional data for filterFormula.
	 * @return Entered product group as a Map.
	 * <pre>
	 * [
	 *   "productFieldName"  : "attribute3",
	 *   "productFieldLabel" : "Business Unit",
	 *   "productFieldValue" : "MeatBall"
	 * ]
	 * </pre>
	 */
	Object productGroupEntry(String entryName, String filterFormulaName, Object filterFormulaParam);

	/**
	 * Creates an input parameter that lets the user select a group of customers and returns the selection as a Map.
	 * This is commonly used in PromotionManager and RebateManager.
	 * It is rendered as a customer group picker widget.
	 *
	 * @param entryName Name of the user entry parameter.
	 * @return Entered customer group as a Map.
	 * <pre>
	 * [
	 *   "customerFieldName" : "attribute3",
	 *   "customerFieldLabel" : "Customer Type",
	 *   "customerFieldValue" : "Restaurant"
	 * ]
	 * </pre>
	 * @see #productGroupEntry(String...)
	 */
	Object customerGroupEntry(String... entryName);

	/**
	 * Creates an input parameter that lets the user select a group of customers and returns the selection as a Map.
	 * This is commonly used in PromotionManager and RebateManager.
	 * It is rendered as a customer group picker widget.
	 * <p>
	 * Only PA searchable Customer Extensions will be available.
	 *
	 * @param entryName Name of the user entry parameter.
	 * @return Entered customer group as a Map.
	 * <pre>
	 * [
	 *   "customerFieldName" : "attribute3",
	 *   "customerFieldLabel" : "Customer Type",
	 *   "customerFieldValue" : "Restaurant"
	 * ]
	 * </pre>
	 * @see #productGroupEntry(String...)
	 */
	Object datamartCustomerGroupEntry(String... entryName);

	/**
	 * Creates an input parameter that lets the user select a group of customers and returns the selection as a Map.
	 * This is commonly used in PromotionManager and RebateManager.
	 * It is rendered as a customer group picker widget.
	 *
	 * @param entryName Name of the user entry parameter.
	 * @param filterFormulaName  Unique name of a formula that will be triggered when the customer picker opens.
	 * @param filterFormulaParam Additional data for filterFormula (e.g. agreement's typedId,...).
	 * @return Entered customer group as a Map.
	 * <pre>
	 * [
	 *   "customerFieldName" : "attribute3",
	 *   "customerFieldLabel" : "Customer Type",
	 *   "customerFieldValue" : "Restaurant"
	 * ]
	 * </pre>
	 * @see #productGroupEntry(String...)
	 */
	Object customerGroupEntry(String entryName, String filterFormulaName, Object filterFormulaParam);

	/**
	 * Creates an input parameter that lets the user enter multiple discounts/surcharges/tiers by amount.
	 * This is commonly used in PromotionManager and RebateManager.
	 * It is rendered as a multi-tier entry widget. Every line holds an amount and the given discount/surcharge in an ascending or descending order.
	 *
	 * @param entryName  Name of the user entry parameter.
	 * @param arguments Two value hints: sort type (can be either "ASC" which is the default or "DESC") and order validation type
	 *               (can be "VALIDATE" or "NO_VALIDATION").
	 * @return Values entered by the user as a TieredValue object.
	 */
	Object multiTierEntry(String entryName, String... arguments);

	/**
	 * Returns a value selected by the user from a list displayed in drop-down input parameter.
	 * <p><b>Side effect</b>: During the Syntax Check mode, this function triggers creation of a drop-down
	 * input parameter with the given list of options.
	 *
	 * <p>
	 * Example 1:
	 * <blockquote>
	 * <pre>
	 * return api.option("Currency", ["EUR", "USD"])
	 * </pre>
	 * </blockquote>
	 *
	 * <p>
	 * Example 2:
	 * <blockquote>
	 * <pre>
	 * def salesOrgs = api.findLookupTableValues("SalesOrg").collect{
	 *   it.name
	 * }
	 *
	 * return api.option("SalesOrg", salesOrgs)
	 * </pre>
	 * </blockquote>
	 *
	 * @param entryName Name of the user entry parameter.
	 * @param options   Arbitrary number of options that should be displayed in the drop-down list.
	 * @return Value selected by the user. During Syntax Check execution mode, it returns mock data.
	 * @see #options(String, Object...)
	 */
	Object option(String entryName, Object... options);

    /**
	 * Creates an input parameter that lets the user select a single value from predefined options.
	 * It is rendered as a drop-down list of possible options.
	 * You can add custom (display-only) labels.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def salesOrgMap = api.findLookupTableValues("SalesOrg").collectEntries{
	 *   [(it.name) : it.name]
	 * }
	 *
	 * return api.option("SalesOrg", salesOrgMap.keySet() as List, salesOrgMap)
	 * </pre>
	 * </blockquote>
     *
     * @param entryName Name of the user entry parameter.
     * @param options   Arbitrary number of options that should be displayed in the drop-down list.
     * @param labels    Map [value : label] of labels that should be shown in the UI. If null or empty, it will behave the same as api.option() without a label definition.
     * @return User entered value.
	 * @see #options(String, List, Map)
     */
    Object option(String entryName, List<Object> options, Map<String, Object> labels);

    /**
     * Creates an input parameter that lets the user select multiple values from predefined options.
	 * It is rendered as a drop-down list of possible options. Each selected value appears at the top.
     *
     * @param entryName Name of the user entry parameter.
     * @param options   Arbitrary number of options that should be displayed in the drop-down list.
     * @return List of user selected values.
	 * @see #option(String, Object...)
     */
    Object options(String entryName, Object... options);

    /**
     * Creates an input parameter that lets the user select multiple values from predefined options.
	 * It is rendered as a drop-down list of possible options. Each selected value appears at the top.
     * You can add custom (display-only) labels.
     *
     * @param entryName Name of the user entry parameter.
     * @param options   Arbitrary number of options that should be displayed in the drop-down list.
     * @param labels    Map [value : label] of labels that should be shown in the UI. If null or empty, it will behave the same as {@link #option(String, Object...)} without a label definition.
     * @return List of user selected values.
     */
    Object options(String entryName, List<Object> options, Map<String, Object> labels);

	/**
	 * Creates an input parameter that lets the user build a filter for a table of objects of the type defined by the {@code typeCode} parameter.
	 * <p>
	 * It is rendered as a filter builder widget allowing to construct a filter (e.g. to define filter for Products, set the {@code typeCode} to {@code P}).
	 * <p>
	 * The filter object can be used in other functions, like {@link #find(String, Filter...)}, {@link #stream(String, String, Filter...)}, {@link #count(String, Filter...)} etc.
	 *
	 * @param entryName Name of the user entry parameter.
	 * @param typeCode  Type code string of the type of the object for which the filter is set. Currently, the supported type codes are "P" (Product data) and "C" (Customer data).
	 * @return Filter object defined by the user.
	 */
	Object filterBuilderUserEntry(String entryName, String typeCode);

	/**
	 * Creates an input parameter that lets the user build a filter for a given Datamart.
	 * <p>
	 * The filter can be used in {@link net.pricefx.formulaengine.DatamartContext.Query#where(Filter...)}
	 * of the {@link PublicGroovyAPI#datamartQuery(Object...)}.
	 * <p>
	 * It will render a FilterBuilder widget entry allowing to construct a filter object filtering on the specified Data Source.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def filter1 = datamartFilterBuilderUserEntry("ProductGroup", "Transactions_DM")
	 *
	 * def dimFilterList = ["CustomerId", "ProductId"]
	 * def filter2 = datamartFilterBuilderUserEntry("ProductGroup", "Transactions_DM", dimFilterList)
	 *
	 * def dimFilterMap = ["CustomerId" : "CD-00155", "ProductId" : "MB-0005"]
	 * def filter3 = datamartFilterBuilderUserEntry("ProductGroup", "Transactions_DM", dimFilterMap, Filter.equal("Country", "DE"))
	 * </pre>
	 * </blockquote>
	 * @param entryName    Name of the user entry parameter.
	 * @param source       Name of the source for which the filter is set.  The resolution of the source is as follows:
     * <ol>
     *                  <li> any Datamart - by typedId, sourceName, uniqueName or label (uniqueName and label for backwards compatibility) </li>
     *                  <li> any FieldCollection - by typedId or sourceName </li>
     * </ol>
	 * @param args There can be up to 2 arguments: one can be a Map or List and the other a Filter. The order does not matter.
	 *             If the argument is a Map, each pair [fieldName : value] will be used as a filter.
	 * @return Filter object to the given source.
	 */
	Object datamartFilterBuilderUserEntry(String entryName, String source, Object... args);

	/**
	 * Looks up the result price (considering a manual override) of an SKU being calculated (SKU in the context) in a price list of the given name.
	 * The price list line has to be approved and the targetDate of the price list has to match the calculation targetDate or be before it.
	 * If multiple price lists are found, the one with the targetDate closer to the calculation targetDate is chosen.
	 *
	 * @param listName Name of the price list to be searched.
	 * @return {@code resultPrice} of the price list record for the SKU.
	 */
	Object pricelist(String listName);

	/**
	 * Looks up the value of the attribute given by {@code attributeName} (considering a manual override) of an SKU being
	 * calculated (SKU in the context) in a price list of the given name.
	 * The price list line has to be approved and the targetDate of the price list has to be prior to the calculation targetDate.
	 * If multiple price lists are found, the one with the targetDate closer to the calculation targetDate is used.
	 *
	 * @param listName      Name of the price list.
	 * @param attributeName Name of the attribute to retrieve.
	 * @return Value of the {@code attributeName} column of the price list record.
	 * @see #pricelistItem(String, String)
	 */
	Object pricelist(String listName, String attributeName);

	/**
	 * Looks up the result price (considering a manual override) of a given SKU in a price list of the given name.
	 * The price list line has to be approved and the targetDate of the price list has to be prior to the calculation targetDate.
	 * If there are multiple price lists of the same name, the one with the targetDate closer to the calculation targetDate is chosen.
	 *
	 * @param listName      Name of the price list.
	 * @param attributeName Name of the attribute to retrieve.
	 * @param sku           SKU to use.
	 * @return Value of the {@code attributeName} column of the price list record.
	 * @see #pricelistItem(String, String)
	 */
	Object pricelist(String listName, String attributeName, String sku);

	/**
	 * Looks up the attribute defined by {@code attributeName} (considering a manual override) of the given SKU in a matrix price list of the given name.
	 * The price list line has to be approved and the targetDate of the price list has to be prior to the calculation targetDate.
	 * If there are multiple price lists of the same name, the one with the targetDate closer to the calculation targetDate is chosen.
	 *
	 * @param listName      Name of the price list.
	 * @param attributeName Name of the attribute to retrieve.
	 * @param sku           SKU to use.
	 * @param key2          key2 to use.
	 * @return Value of the {@code attributeName} column of the matrix price list record.
	 * @see #pricelistItem(String, String, String)
	 */
	Object pricelist(String listName, String attributeName, String sku, String key2);

	/**
	 * Looks up the price list item record (PLI or XPLI) of an SKU being calculated in a price list of the given name.
	 * The price list line has to be approved and the targetDate of the price list has to be prior to the calculation targetDate.
	 * If there are multiple price lists of the same name, the one with the targetDate closer to the calculation targetDate is chosen.
	 *
	 * @param listName Name of the price list.
	 * @return Entire price list item record (PLI or XPLI).
	 * @see #pricelist(String)
	 */
	Object pricelistItem(String listName);

	/**
	 * Looks up the price list item record (PLI) of the given SKU in a price list of a given name.
	 * The price list line has to be approved and the targetDate of the price list has to be prior to the calculation targetDate.
	 * If there are multiple price lists of the same name, the one with the targetDate closer to the calculation targetDate is chosen.
	 *
	 * @param listName Name of the price list.
	 * @param sku      SKU to use.
	 * @return Entire price list item record (PLI).
	 * @see #pricelist(String, String, String)
	 */
	Object pricelistItem(String listName, String sku);

	/**
	 * Looks up the matrix price list item record (XPLI) of the given SKU in a matrix price list of the given name.
	 * The price list line has to be approved and the targetDate of the price list has to be prior to the calculation targetDate.
	 * If there are multiple price lists of the same name, the one with the targetDate closer to the calculation targetDate is chosen.
	 *
	 * @param listName Name of the price list.
	 * @param sku      SKU to use.
	 * @param key2     key2 to use.
	 * @return Entire matrix price list item record (XPLI).
	 * @see #pricelist(String, String, String, String)
	 */
	Object pricelistItem(String listName, String sku, String key2);

	/**
	 * Looks up the matrix price list item record (XPLI) of the given SKU in a matrix price list of the given name.
	 * The price list line has to be approved and the targetDate of the price list has to be prior to the calculation targetDate.
	 * If there are multiple price lists of the same name, the one with the targetDate closer to the calculation targetDate is chosen.
	 *
	 * @param listName Name of the price list.
	 * @param sku      SKU to use.
	 * @param key2     key2 to use.
	 * @param targetDate targetDate for the search.
	 * @return Entire matrix price list item record (XPLI).
	 * @see #pricelist(String, String, String, String)
	 */
	Object pricelistItem(String listName, String sku, String key2, Date targetDate);

	/**
	 * Similar to the {@link #pricelistItem(String)} function. However, it optimizes the subsequent processing by providing the results as a map directly.
	 *
	 * @param listName Name of the price list. Target date and SKU matching applies.
	 * @return Price list object.
	 * @see #pricelistItem(String)
	 */
	Map<String, Object> getPricelistItem(String listName);

	/**
	 * Accesses and retrieves values from a rolled-up bill of materials for the SKU being calculated.
	 * All quantities are rolled up and summed.
	 *
	 * @return List with the summed up quantities.
	 */
	Object bom();

	/**
	 * Accesses and retrieves values from a rolled-up bill of materials for the SKU being calculated.
	 * The summed up quantities are multiplied with a factor that is retrieved by a lookup by the material name from the provided parameter,
	 * for example, a material-cost table.
	 *
	 * @param lookupTableName Parameter table name.
	 * @return List with the summed up and multiplied quantities.
	 */
	Object bom(String lookupTableName);

	/**
	 * Accesses and retrieves values from a rolled-up bill of materials for the SKU being calculated.
	 * Only BoM records with the specified category are considered.
	 * The summed up quantities are multiplied with a factor that is retrieved by a lookup by the material name from the provided parameter,
	 * for example, a material-cost table.
	 *
	 * @param lookupTableName Parameter table name.
	 * @param categoryName    Category name to filter on.
	 * @return List with the summed up and multiplied quantities.
	 */
	Object bom(String lookupTableName, String categoryName);

	/**
	 * Accesses and retrieves the values from a rolled-up bill of materials for the SKU being calculated.
	 * Only BoM records with the specified category and material are considered.
	 * The summed up quantities are multiplied with a factor that is retrieved by a lookup by the material name from the provided parameter,
	 * for example, a material-cost table.
	 *
	 * @param lookupTableName Parameter table name.
	 * @param categoryName    Category name to filter on.
	 * @param materialName    Material to filter on.
	 * @return List with the summed up and multiplied quantities.
	 */
	Object bom(String lookupTableName, String categoryName, String materialName);

	/**
	 * Rolls up the current SKU's BoM and returns the rolled up BoM records (not just the quantities).
	 *
	 * @return List of BoM records
	 */
	Object bomList();

	/**
	 * Creates an input parameter that lets the user select a user from the list of all users.
	 * It is rendered as a user picker dialog / select list.
	 *
	 * @param entryName Name of the entry parameter
	 * @return Login name of the selected user or {@code null}
	 */
	Object anyUser(String entryName);

	/**
	 * Returns a value from the current user's master data.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * // retrieves the email of the user
	 * def email = api.user("email")
	 * </pre>
	 * </blockquote>
	 *
	 * @param attributeName Attribute to return.
	 * @return User's attribute value
	 */
	Object user(String attributeName);

	/**
	 * Returns the entire master data object of the current user.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * // retrieves the email of the user
	 * def email = api.user()?.email
	 * </pre>
	 * </blockquote>
	 *
	 * @return user object (U)
	 */
	Object user();

	/**
	 * Retrieves the value of the input parameter of the given name.
	 * It only reads the value, it does not trigger rendering any input widget (i.e. no side effects).
	 *
	 * @param inputName Input parameter name.
	 * @return Value entered by the user in the input parameter. In the Syntax Check execution mode, it returns mock data.
	 * @see #userEntry(String)
	 */
	Object input(String inputName);

	/**
	 * Creates an input parameter that lets the user select a matrix of values.
	 * It renders as a grid-style input widget with the specified columns.
	 * The result will be a list (= rows) of maps (= row's columns). The map attribute names are the column names.
	 * <p>
	 * Make sure the column names are valid JSON identifiers (spaces are ok but try to avoid special chars).
	 *
	 * @param inputMatrixName Name of the parameter.
	 * @param columnNames     Column names/labels.
	 * @return List (= rows) of maps (= row's columns).
	 */
	Object inputMatrix(String inputMatrixName, String... columnNames);

    /**
     * Creates an input parameter that lets the user select a matrix of values.
	 * It renders as a grid-style input widget with the specified columns.
     * The result will be a list (= rows) of maps (= row's columns). The map attribute names are the column names.
	 * <p>
     * Make sure the column names are valid JSON identifiers (spaces are ok but try to avoid special chars).
     *
     * @param inputMatrixName Name of the parameter.
     * @param paramConfig     Map of the configuration settings with the following options (key, value):
     *                        <ul>
     *                          <li>'canModifyRows' - false/true (the user can add and delete a row)</li>
     *                          <li>'readOnlyColumns' - list of columns (names) to be uneditable </li>
     *                          <li>'noCellRefresh' - false/true (do not update Configurator on each edit)</li>
     *                          <li>'defaultHeight' - number which specifies the default table height</li>
     *                          <li>'fixTableHeight' - false/true (to preserve the default height of the table)</li>
     *                          <li>'fitFieldWidths' - false/true (to determine if fields should autofit their widths to titles or content)</li>
     *                        </ul>
     * @param columns     Column names/labels.
     * @return List (= rows) of maps (= row's columns).
     */
	Object inputMatrix(String inputMatrixName, Map<String,Object> paramConfig, String...columns);

	/**
	 * Creates an input parameter that lets the user pick a file.
	 * It renders as a file input widget that allows to use data from an XLSX file in a logic.
	 * It works only for entities with attachments (i.e. Q/CT/RBA) which need to be saved first.
	 * @param inputName Name of the parameter.
	 * @return Handle that uniquely identifies the binary and its version that was assigned to the input.
	 */
	Object parsableInputFile(String inputName);

	/**
	 * If a parsableInputFile has a value, this function opens the file and parses it to basic Groovy data structures.
	 * @param inputName Name of the parameter.
	 * @return [data: [sheetName: [row1col1, row1col2, ...], [row2col1, row2col2, ...], ...]
	 */
	Object parsableInputFileData(String inputName);

	/**
	 * If a parsableInputFile has a value, this function opens the file and parses it to basic Groovy data structures.
	 * @param fileHandle Value of the parameter.
	 * @return [data: [sheetName: [row1col1, row1col2, ...], [row2col1, row2col2, ...], ...]
	 */
	Object parsableInputFileDataFromHandle(String fileHandle);

	/**
	 * Creates a filter object that can be used in other API functions. The operator is always "equals", i.e. property = value.
     * <p>
     * Example:
     * <blockquote>
	 * <pre>
	 * def cas = api.find("CA",
	 * 		api.filter("assignmentId", mplId),
	 * 		api.filter("assignmentType", "MPL")
	 * )
	 * </pre>
	 * </blockquote>
	 *
	 * @param property Name of the property to filter on.
	 * @param value    Specified value.
	 * @return Filter object
	 * @deprecated Use {@link Filter#equal(String, Object)} instead
	 */
	Object filter(String property, Object value);

	/**
	 * Constructs a filter object from the map representation of that filter (from the "wire format").
	 * The map representation is the way the filter is described in the JSON requests between the UI and the server.
	 *
	 * @param filterMap Filter in the form of nested maps.
	 * @return Filter object
	 */
	Filter filterFromMap(Map<String, Object> filterMap);

	/**
	 * Creates a filter object that can be used in other API functions.
	 * <p>
	 * Note: This method is, to an extent, dynamic. Dynamic in the sense that if the {@code opString} does not actually represent
	 * a valid operator (e.g. "="), the passed value will then be treated as the lower bound and the {@code value} as the upper
	 * bound. In other words, the resulting filter will look like this:
	 * {@code Filter.and(Filter.greaterOrEqual(property, opString), Filter.lessThan(property, value))}.
	 * If either parameter is {@code null}, its part is skipped.
     *
     * <p>Example:
     * <blockquote>
	 * <pre>{@code
	 * datamartQuery.where(api.filter("PricingDate", ">", targetDate.minus(365).format("yyyy-MM-dd")))
	 * }</pre>
	 * </blockquote>
     *
	 * @param property Name of the property to filter on.
	 * @param opString One of these: "=", "&lt;", "&lt;=", "&gt;", "&gt;=", "&lt;&gt;", "~". If different, it gets treated as the lower bound value; see the note above.
	 * @param value    Specified value (or an upper bound).
	 * @return Filter object
	 * @see Filter#equal(String, Object)
	 * @see Filter#lessThan(String, Object)
	 * @see Filter#lessOrEqual(String, Object)
	 * @see Filter#greaterOrEqual(String, Object)
	 * @see Filter#greaterThan(String, Object)
	 * @see Filter#notEqual(String, Object)
	 * @see Filter#like(String, String)
	 * @see Filter#and(Filter...)
	 * @see Filter#or(Filter...)
	 * @see Filter#in(String, Object...)
	 * @see Filter#notIn(String, Object...)
	 * @deprecated Use the {@link Filter} functions instead
	 */
	Object filter(String property, Object opString, Object value);

	/**
	 * Creates a filter to be applied to a search on a related object type, limiting the result to objects
	 * of the related type that has a link to the given customer.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * 		def filter = api.customerToRelatedObjectsFilter("RBA", "104083")	// customerId=104083
	 * 		def rbas = api.find("RBA", 0, 0, null, filter)
	 * 		api.trace("rbas for customer #104083", null, rbas)
	 * </pre>
	 * </blockquote>
	 *
	 * @param relatedObjectTypeCode	TypeCode of the related object type. The possible values are:
	 * <ul>
	 * 		<li>'PR':    PriceRecord</li>
	 * 		<li>'CT':    Contract</li>
	 * 		<li>'CTLI':  ContractLineItem</li>
	 * 		<li>'RBA':   RebateAgreement</li>
	 * 		<li>'RBALI': RebateAgreementLineItem</li>
	 * </ul>
	 * @param customerId Customer.id to filter on.
	 * @return Filter object
	 */
	Filter customerToRelatedObjectsFilter(String relatedObjectTypeCode, String customerId);

	/**
	 * Creates a filter to be applied to a customer search, limiting the results to customers that are
	 * linked to the given related object (defined by its typedId).
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def ras = api.find("RBA", 0, 1, null, null)
	 * ras.each {
	 * 		def filter = api.relatedObjectToCustomersFilter(it.typedId)
	 * 		def customers = api.find("C", 0, 0, null, filter)
	 * 		...
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @param relatedObjectTypedId	TypedId of the related object. Its type needs to be one of:
	 * <ul>
	 * 		<li>'PR' - PriceRecord</li>
	 * 		<li>'CT' - Contract</li>
	 * 		<li>'CTLI' - ContractLineItem</li>
	 * 		<li>'RBA' - RebateAgreement</li>
	 * 		<li>'RBALI' - RebateAgreementLineItem</li>
	 * </ul>
	 * @return Filter object
	 */
	Filter relatedObjectToCustomersFilter(String relatedObjectTypedId);

	/**
	 * Creates a filter to be applied to a search on a related object type, limiting the results
	 * to objects of the related type that have a link to the given product.
	 <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def ras = api.find("RBA", 0, 1, null, null)
	 * ras.each {
	 * 		def raId = it.typedId.tokenize(".")[0] as Long
	 *   	def raFilter = api.relatedObjectToCustomersFilter("RBA", raId)
	 * 		def customers = api.find("C", 0, 0, null, filter)
	 * 		...
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @param relatedObjectTypeCode	TypeCode of the related object type. Possible values are:
	 * <ul>
	 * 		<li>'PR' - PriceRecord</li>
	 * 		<li>'CT' - Contract</li>
	 * 		<li>'CTLI' - ConttractLineItem</li>
	 * 		<li>'RBA' - RebateAgreement</li>
	 * 		<li>'RBALI' - RebateAgreementLineItem</li>
	 * </ul>
	 * @param sku Product.id to filter on.
	 * @return Filter object
	 */
	Filter productToRelatedObjectsFilter(String relatedObjectTypeCode, String sku);

	/**
	 * Creates a filter to be applied to a product search, limiting the result to products
	 * that are linked to the given related object (defined by its typedId).
	 *
	 * @param relatedObjectTypedId	TypedId of the related object. Its type needs to be one of:
	 * <ul>
	 * 		<li>'PR' - PriceRecord</li>
	 * 		<li>'CT' - Contract</li>
	 * 		<li>'CTLI' - ContractLineItem</li>
	 * 		<li>'RBA' - RebateAgreement</li>
	 * 		<li>'RBALI' - RebateAgreementLineItem</li>
	 * </ul>
	 * @return Filter object
	 */
	Filter relatedObjectToProductsFilter(String relatedObjectTypedId);

	/**
	 * Checks if the current user is a member of the given user group.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def user = api.user("loginName")
	 * if (api.isUserInGroup("Manager", user)) {
	 *     ...
	 * }
	 * </pre>
	 * </blockquote>
	 * @param userGroupName Group name.
	 * @param loginUserName Login name of the user.
	 * @return True or false
	 */
	boolean isUserInGroup(String userGroupName, String loginUserName);

	/**
	 * Generates a trace message that can be used during logic testing.
	 * The Trace window has 3 columns (FunctionName, Parameters, Result) which you can use for your debugging messages.
	 * During the regular logic execution the trace messages are ignored.
	 * The parameters specify the output of the trace.
	 * <p>Note: Consider commenting out the trace messages before deploying to production.</p>
	 * <p>Performance consideration: Be carefull when using the api.trace() in a loop. If your logic (during testing)
	 * produces a huge list of trace messages on the backend, then this big list must be transferred over the internet
	 * to your local machine (when using Studio) and it can take significant time.</p>
     *
	 * <p>Example:
	 * <blockquote><pre>
	 *     def listPrice = ...
	 *     api.trace("listPrice", out.Currency, listPrice)
	 * </pre></blockquote>
	 *
* @param functionName Function name. Regardless of the name, it can be any text you need to show during testing. Usually, if you use trace() for an output of a variable, you put there the variable name. (Originally this was used as a name of the function which outputted the value, so that is the origin of the name.)
* @param parameters   Parameters to trace. Regardless of the name, it can be any text you need to show during testing. (Originally this was used to display values of parameters supplied into a function, hence the name.)
* @param result  Result of the trace. Usually it is the value which you are tracing during testing. You can supply any value you would like to see in the Trace window. (Originally this was used to show a result of a function call, hence the name.)
	 *
	 * @see #trace(Object)
	 * @see #trace(String, Object)
	 */
	void trace(String functionName, String parameters, Object result);

	/**
	 * See {@link #trace(String, String, Object)} for a description.
	 *
	 * <p>Example:
	 * <blockquote><pre>
	 *     def salesOrg = ...
	 *     api.trace("salesOrg", salesOrg)
	 * </pre></blockquote>
	 *
	 * @param functionName Function name. Regardless of the name, it can be any text you need to show during testing. Usually, if you use trace() for an output of a variable, you put there the variable name. (Originally this was used as a name of the function which outputted the value, so that is the origin of the name.)
	 * @param result       Result of the trace. Usually it is the value which you are tracing during testing. You can supply any value you would like to see in the Trace window. (Originally this was used to show a result of a function call, hence the name.)
	 *
	 * @see #trace(String, String, Object)
	 * @see #trace(Object)
	 */
	void trace(String functionName, Object result);

	/**
	 * See {@link #trace(String, String, Object)} for a description.
	 *
	 * <p>Example:
	 * <blockquote><pre>
	 *     def price = ...
	 *     api.trace(price)
	 * </pre></blockquote>
	 *
	 * @param result       Result of the trace. Usually it is the value which you are tracing during testing. You can supply any value you would like to see in the Trace window. (Originally this was used to show a result of a function call, hence the name.)
	 *
	 * @see #trace(String, String, Object)
	 * @see #trace(String, Object)
	 */
	void trace(Object result);

	/**
	 * This function is meant to be used by the Pricefx application internally only. You should use {@link #logInfo(String, Object)} or {@link #logWarn(String, Object)}.
	 * <p>
	 * Logs a DEBUG message to the system error log which can be displayed in Tools &gt; Logs &gt; View Log.
	 * The log message can contain placeholders like {} which are then filled in dynamically with the argument's values.
	 * <b>Note</b>: To enable this function, you need to select the option "Allow Groovy logging to system log" in Tools &gt; Configuration &gt; General Settings.
	 *
	 * @param msg  Log message to print.
	 * @param arg1 Arguments for the log message.
	 * @see #logInfo(String, Object)
	 * @see #logWarn(String, Object)
	 */
	void log(String msg, Object arg1);

	/**
	 * This function can be used by the Pricefx Support team only. You should use {@link #logInfo(Object)} or {@link #logWarn(Object)}.
	 * <p>
	 * Logs a DEBUG message to the system error log which can be displayed in menu Tools &gt; Logs &gt; View Log.
	 * The log message can contain placeholders like {} which are then filled in dynamically with the argument's values.
	 * <b>Note</b>: To enable this, you need to select the option "Allow Groovy logging to system log" in Tools &gt; Configuration &gt; General Settings.
	 *
	 * @param msg  Message accompanying the exception.
	 * @see #logInfo(Object)
	 * @see #logWarn(Object)
	 */
	void log(Object msg);

	/**
	 * Logs an INFO warning message to the system error log which can be displayed in Tools &gt; Logs &gt; View Log.
	 * The log message can contain placeholders like {} which are then filled in dynamically with the argument's values.
	 * <b>Note</b>: To enable this, you need to select the option "Allow Groovy logging to system log" in Tools &gt; Configuration &gt; General Settings.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def sku = api.product("sku")
	 * api.markItemDirty()
	 * api.logInfo("SKU marked as dirty", sku)
	 * </pre>
	 * </blockquote>
	 *
	 * @param msg  Log message to print.
	 * @param arg1 Arguments for the log message.
	 * @see #log(String, Object)
	 * @see #logWarn(String, Object)
	 */
	void logInfo(String msg, Object arg1);

	/**
	 * Logs an INFO message to the system error log which can be displayed in Tools &gt; Logs &gt; View Log.
	 * The log message can contain placeholders like {} which are then filled in dynamically with the argument's values.
	 * <b>Note</b>: To enable this, you need to select the option "Allow Groovy logging to system log" in Tools &gt; Configuration &gt; General Settings.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * api.logInfo("No items to be calculated")
	 * </pre>
	 * </blockquote>
	 *
	 * @param msg Log message to print.
	 * @see #log(Object)
	 * @see #logWarn(Object)
	 */
	void logInfo(Object msg);

	/**
	 * Logs a WARN message to the system error log which can be displayed in Tools &gt; Logs &gt; View Log.
	 * The log message can contain placeholders like {} which are then filled in dynamically with the argument's values.
	 * <b>Note</b>: To enable this, you need to select the option "Allow Groovy logging to system log" in Tools &gt; Configuration &gt; General Settings.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def sku = api.product("sku")
	 *
	 * if (checkItem()) {
	 *      api.logWarn("Incorrect or missing 'sku' parameter", sku)
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @param msg  Log message to print.
	 * @param arg1 Arguments for the log message.
	 * @see #log(String, Object)
	 * @see #logInfo(String, Object)
	 */
	void logWarn(String msg, Object arg1);

	/**
	 * Logs a WARN message to the system error log which can be displayed in Tools &gt; Logs &gt; View Log.
	 * The log message can contain placeholders like {} which are then filled in dynamically with the argument's values.
	 * <b>Note</b>: To enable this, you need to select the option "Allow Groovy logging to system log" in Tools &gt; Configuration &gt; General Settings.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * api.logWarn("SKU cannot be null")
	 * </pre>
	 * </blockquote>
	 *
	 * @param msg  Message accompanying the exception.
	 * @see #log(Object)
	 * @see #logInfo(Object)
	 */
	void logWarn(Object msg);

	/**
	 * Retrieves an already generated (in the previous logic code) input parameter in the form of a {@link ContextParameter} object.
	 * This object can be used to set a label and a default value of an input, if it is required, etc.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>{@code
	 * def p = api.getParameter("Quantity")        // retrieve the context parameter with the same name as the input
	 * if (p != null && p.getValue() == null) {
	 *   p.setLabel("Required Quantity")           // set the displayed label
	 *   p.setRequired(true)                       // set the mandatory hint
	 *   p.setReadOnly(false)                      // set the read only flag
	 *   p.setValue(1)                             // set the default value with which the input is pre-populated
	 * }
	 * return qty ?: 1                             // if user doesn't specify a value, return 1
	 * }</pre>
	 * </blockquote>
	 *
	 * @param parameterName Name of the input parameter to retrieve.
	 * @return Parameter object or {@code null}.
	 */
	ContextParameter getParameter(String parameterName);

	/**
	 * In case of list processing (price lists, price grids), this method allows an easy access to values from the same list.
	 * It searches for the result record in the same list with the provided SKU.
	 * <p>
	 * Note: If used in a context that has a secondary key, only the first object is returned.
	 *
	 * @param sku SKU to look for.
	 * @return Result record (entire record)
	 */
	Map<String, Object> currentContext(String sku);

	/**
	 * In case of list processing (price lists, price grids), this method allows an easy access to values from the same list.
	 * It searches for the result record in the same list with the provided SKU.
	 *
	 * @param sku  SKU to look for.
	 * @param key2 key2 to look for.
	 * @return result record (entire record)
	 */
	Map<String, Object> currentContext(String sku, String key2);

	/**
	 * In case of (chained) price lists, this method allows for an easy retrieval of records from the previous list.
	 * <p>
	 * Note: If used in a context that has a secondary key, only the first object is returned.
	 *
	 * @param sku SKU to look for.
	 * @return Result record (entire record)
	 */
	Map<String, Object> previousContext(String sku);

	/**
	 * In case of (chained) price lists, this method allows for an easy retrieval of records from the previous list.
	 *
	 * @param sku  SKU to look for.
	 * @param key2 key2 to look for.
	 * @return Result record (entire record)
	 */
	Map<String, Object> previousContext(String sku, String key2);

	/**
	 * Name of the current context, e.g. price list name, quote name etc.
	 *
	 * @return Name of a context.
	 */
	String contextName();

	/**
	 * The type (type code string) of the current context.
	 * @return Type code string
	 */
	String contextType();

	/**
	 * The typedId of the current context.
	 * @return typedId string
	 */
	String contextTypedId();

	/**
	 * Provides an optimized way to get to certain object attributes. The sandbox automatically converts "real" server objects into
	 * a map representation when the object is accessed. This is for safety (sandboxing) reasons. This process converts
	 * all object attributes. If only one specific attribute is of interest, this method shortcuts that as the underlying object is not exposed
	 * (and thus does not force the sandbox into conversion).
	 *
	 * @param object Object whose value to retrieve.
	 * @param path   Attribute name.
	 * @return Value of the attribute
	 */
	Object getProperty(Object object, String path);

	/**
	 * Gets object attributes by elementName. It works only for objects that have meta attributes which contain the element name.
	 * These are e.g. PLI, SIMI, PGI, MPLI.
	 *
	 * @param object      Object whose value should be retrieved.
	 * @param elementName elementName value.
	 * @return Value of the attribute or {@code null} if an invalid object or no matching meta found.
	 */
	Object getPropertyByElementName(Object object, String elementName);

	/**
	 * In case of a list processing operation (for price lists, price grids), this method gets the currently worked on record, for example:
	 * <ul>
	 *     <li>For price calculations, it returns the price list item or price grid item or quote item.</li>
	 *     <li>For Calculated Field Sets, it returns the row being processed, like Product/Customer/ProductExtension/PriceParameter row.</li>
	 * </ul>
	 * <p><strong>Important</strong>: The price list line item is initially (during 1st pass) created only after the first logic execution
     * (i.e. the PLI is created from the logic results). That means that during the first logic execution,
	 * the PLI does not exist yet and the method will return {@code null}.
     * <p>
	 * Example - in a price list logic:
	 * <blockquote>
	 * <pre>
     * def currentItem = api.currentItem()
     * def currentSku = currentItem != null ? currentItem.attribute20 : api.product("attribute20")
	 * </pre>
	 * </blockquote>
     * <p>
	 * Example - in a Calculated Field Set:
	 * <blockquote>
	 * <pre>
     * def cost = api.currentItem()?.attribute1
     * def exchangeRate = 1.1 // fixed rate for sample purposes
     * return cost != null ? (cost * exchangeRate) : null
	 * </pre>
	 * </blockquote>
	 *
	 * @return Item object as a Map.
	 *
	 * @see #currentItem(String)
	 * @see #currentItemByElementName(String)
	 */
	Object currentItem();

	/**
	 * In case of a list processing operation (for price lists, price grids), this method gets the currently worked on record's property by elementName.
	 * <p>For example, when you want to get a value of an LPG column which was generated by an element named ListPrice,
	 * you do not need to know that it was stored in the column attribute3,
	 * but you can ask for the value using the name of the element which provided the value.
	 *
     *
	 * @param elementName Element name
	 * @return Item's property value
     *
	 * @see #currentItem(String)
	 */
	Object currentItemByElementName(String elementName);

	/**
	 * The same as {@link #currentItem} but it returns only the value of the given attribute.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def activePrice = api.currentItem("activePrice")
	 * </pre>
	 * </blockquote>
	 *
	 * @param attributeName Attribute to retrieve.
	 * @return Value of the attribute.
	 *
	 * @see #currentItemByElementName(String)
	 */
	Object currentItem(String attributeName);

	/**
	 * Returns a set of SKUs which comprise so-called 'SKU context' of
	 * the current formula execution.
	 * <p>
	 * The form of the SKU context differs, e.g., in case of the
	 * product detail formula executed via price grid details,
	 * the resulting set will contain all SKUs of the left-pane grid.
	 * <p>
	 * Note:
	 * <ul>
	 * <li>If the SKU context is not supported by the current formula
	 * execution, {@code null} is returned.</li>
	 * <li>A maximum of 200 SKUs is returned per one call. If there are more,
	 * {@code startRow} must be used to get all SKUs in the set.</li>
	 * </ul>
	 * @param startRow Starting row of the result set.
	 * @return Set of SKUs or {@code null} if the SKU context is not
	 * available for the current formula execution.
	 */
	Set<Object> contextSkuSet(int startRow);

	/**
	 * Returns the maximum number of records that {@link #find(String, int, int, String, Filter...)} and its variations can return in one call.
	 * The value is set in the {@code formulaEngine.script.findMaxResults} system parameter.
	 * The default value is 2000 records.
	 *
	 * @return Maximum number of records
	 * @see #find(String, int, int, String, Filter...)
	 */
	int getMaxFindResultsLimit();

	/**
	 * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
	 * The filters are ANDed together. A maximum of 200 items is returned.
	 *
	 * @param typeCode Type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}.
	 * @param filters  Optional list of filters (that will evaluated using the AND operator).
	 * @return List of objects represented as Maps.
	 * @see #stream(String, String, Filter...)
	 * @see #count(String, Filter...)
     * @deprecated Instead use {@link #find(String, int, int, String, List, Filter...)} in order to always specify the fields to be returned (for performance reasons).
     */
	List<Object> find(String typeCode, Filter... filters);

	/**
	 * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
	 * The filters are ANDed together. A maximum of 200 items is returned.
	 * The start row parameter can be used to page through more than 200 results.
	 * <p>
	 * If you do not need all the columns to be fetched, you should use {@link #find(String, int, int, String, List, Filter...)} instead (because of performance reasons).
	 *
	 * @param typeCode Type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}.
	 * @param filters  Optional list of filters (that will be evaluated using the AND operator).
	 * @param startRow Starting row of the result set.
	 * @return List of objects represented as Maps.
	 * @see #stream(String, String, Filter...)
	 * @see #count(String, Filter...)
     * @deprecated Instead use {@link #find(String, int, int, String, List, Filter...)} where you should always specify the fields to be returned.
	 */
	List<Object> find(String typeCode, int startRow, Filter... filters);

	/**
	 * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
	 * The filters are ANDed together. A maximum of 200 items is returned.
	 * The start row parameter can be used to page through more than 200 results.
	 * <p>
* If you do not need all the columns to be fetched, you should use {@link #find(String, int, int, String, List, Filter...)} instead (because of performance reasons).
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def sku = api.product("sku")
	 *
	 * def filters = [
	 *     Filter.equal("name", "Competition"),
	 *     Filter.equal("sku", sku),
	 * ]
	 *
	 * // sort by competitor's price ASC
	 * def competitors = api.find("PX", 0, "attribute1", *filters)
	 *
	 * competitors.each {
	 *     ...
	 * }
	 * </pre>
	 * </blockquote>
	 * @param typeCode Type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}.
	 * @param startRow Starting row of the result set.
	 * @param sortBy   String of fields/attributes, separated by a comma, based on which the result set should be sorted. Any field can be prefixed with '-' for descending sort order.
	 * @param filters  Optional list of filters (that will be evaluated using AND operator).
	 * @return List of objects represented as Maps.
	 * @see #stream(String, String, Filter...)
	 * @see #count(String, Filter...)
     * @deprecated Instead use {@link #find(String, int, int, String, List, Filter...)} where you should always specify the fields to be returned.
	 */
	List<Object> find(String typeCode, int startRow, String sortBy, Filter... filters);

	/**
	 * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
	 * The filters are ANDed together. The maximum number of rows retrieved is determined by the formulaEngine.script.findMaxResults system parameter.
	 * The maxRows parameter can only be set to values lower than that.
	 * <p>
     * Example:
     * <blockquote>
	 * <pre>
	 * def sku = api.product("sku")
	 *
	 * def filters = [
	 *     Filter.equal("name", "Competition"),
	 *     Filter.equal("sku", sku),
	 * ]
	 *
	 * // sort by competitor's price ASC
	 * def competitors = api.find("PX", 0, api.getMaxFindResultsLimit(), "-attribute2", *filters)
	 *
	 * competitors.each {
	 *     ...
	 * }
	 * </pre>
	 * </blockquote>
	 * @param typeCode Type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}.
	 * @param startRow Starting row of the result set.
	 * @param maxRows  Maximum number of rows to retrieve. The value can be up to {@link #getMaxFindResultsLimit()}. If set to 0, it will default to 200.
	 * @param sortBy   String of fields/attributes, separated by a comma, based on which the result set should be sorted. Any field can be prefixed with '-' for descending sort order.
	 * @param filters  Optional list of filters (that will be evaluated using the AND operator).
	 * @return List of objects represented as Maps.
	 * @see #getMaxFindResultsLimit()
	 * @see #stream(String, String, Filter...)
	 * @see #count(String, Filter...)
	 */

	List<Object> find(String typeCode, int startRow, int maxRows, String sortBy, Filter... filters);

	/**
	 * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
	 * The filters are ANDed together. The maximum number of rows retrieved is determined by the formulaEngine.script.findMaxResults system parameter.
	 * The maxRows parameter can only be set to values lower than that.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def sku = api.product("sku")
	 *
	 * def filters = [
	 *     Filter.equal("name", "Competition"),
	 *     Filter.equal("sku", sku),
	 * ]
	 *
	 * // sort by competitor's price ASC
	 * def competitors = api.find("PX", 0, api.getMaxFindResultsLimit(), "-attribute2", ["attribute2", "attribute5", "attribute6"], *filters)
	 *
	 * competitors.each {
	 *     ...
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @param typeCode Type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}.
	 * @param startRow Starting row of the result set.
	 * @param maxRows  Maximum number of rows to retrieve. The value can be up to {@link #getMaxFindResultsLimit()}. If set to 0, it will default to 200.
	 * @param sortBy   String of fields/attributes, separated by a comma, based on which the result set should be sorted. Any field can be prefixed with '-' for descending sort order.
	 * @param fields   List of fields to be returned in the result. If you really need all fields, you can set this parameter to null.
	 * @param filters  Optional list of filters (that will be evaluated using the AND operator).
	 * @return List of objects represented as Maps.
	 * @see #getMaxFindResultsLimit()
	 * @see #stream(String, String, List, Filter...)
	 * @see #count(String, Filter...)
	 */
	List<Object> find(String typeCode, int startRow, int maxRows, String sortBy, List<String> fields, Filter... filters);

	/**
	 * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
	 * The filters are ANDed together. The maximum number of rows retrieved is determined by the formulaEngine.script.findMaxResults system parameter.
	 * The maxRows parameter can only be set lower values than that.
	 *
	 * @param typeCode           Type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}.
	 * @param startRow           Starting row of the result set.
	 * @param maxRows            Maximum number of rows to retrieve. The value can be up to {@link #getMaxFindResultsLimit()}. If set to 0, it will default to 200.
	 * @param sortBy             String of fields/attributes, separated by a comma, based on which the result set should be sorted. Any field can be prefixed with '-' for descending sort order.
	 * @param fields             List of fields to be returned in the result. If you really need all fields, you can set this parametr to null.
	 * @param distinctValuesOnly Fetches distinct/all values. The use case here is fetching distinct single column values.
	 * @param filters            Optional list of filters (that will be evaluated using the AND operator).
	 * @return List of objects represented as Maps.
	 * @see #getMaxFindResultsLimit()
	 * @see #stream(String, String, List, boolean, Filter...)
	 * @see #count(String, Filter...)
	 */
	List<Object> find(String typeCode, int startRow, int maxRows, String sortBy, List<String> fields, boolean distinctValuesOnly, Filter... filters);

	/**
	 * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
	 * The filters are ANDed together. The maximum number of rows retrieved is determined by the formulaEngine.script.findMaxResults system parameter.
	 * The maxRows parameter can only be set lower values than that.
	 *
	 * @param typeCode           Type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}.
	 * @param startRow           Starting row of the result set.
	 * @param maxRows            Maximum number of rows to retrieve. The value can be up to {@link #getMaxFindResultsLimit()}. If set to 0, it will default to 200.
	 * @param sortBy             String of fields/attributes, separated by a comma, based on which the result set should be sorted. Any field can be prefixed with '-' for descending sort order.
	 * @param fields             Map of pairs [fieldname : aggregation] holding fields to be returned in the result where the aggregation can be one of these: SUM, AVG, MIN or MAX.
	 * 							 Key of the map is the field name. The value can be a field operator like SUM, AVG, MIN or MAX.
	 * @param distinctValuesOnly Fetches distinct/all values. The use case here is fetching distinct single column values.
	 * @param filters            Optional list of filters (that will be evaluated using the AND operator).
	 * @return List of objects represented as Maps.
	 * @see #getMaxFindResultsLimit()
	 * @see #stream(String, String, Map, boolean, Filter...)
	 * @see #count(String, Filter...)
	 */
	List<Object> find(String typeCode, int startRow, int maxRows, String sortBy, Map<String,String> fields, boolean distinctValuesOnly, Filter ... filters);

	/**
	 * Searches for all types of objects with a freestyle query and iterates over the result set.
	 * The filters are ANDed together.
	 * If you do not need all the columns to be fetched, you should use {@link #stream(String, String, List, Filter...)} instead (because of performance reasons).
	 * <p>
	 * Note: You have to close the iterator after iterating the data using {@link ResultIterator#close()}.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>{@code
	 * def filters = [
	 *      Filter.equal("attribute2", "Active")
	 * ]
	 *
	 * def products = api.stream("P", "-attribute1", *filters)
	 *
	 * products.each { product ->
	 *      ...
	 * }
	 * }</pre>
	 * </blockquote>
	 *
	 * @param typeCode Type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}.
	 * @param sortBy   String of fields/attributes, separated by a comma, based on which the result set should be sorted. Any field can be prefixed with '-' for descending sort order.
	 * @param filters  Optional list of filters (that will be evaluated using the AND operator).
	 * @return Result implements the Iterator interface, allowing for iteration over all records in the scope of the query.
	 * @see #find(String, int, String, Filter...)
	 * @see #count(String, Filter...)
	 */
	ResultIterator stream(String typeCode, String sortBy, Filter... filters);

	/**
	 * Searches for all types of objects with a freestyle query and iterates over the result set.
	 * The filters are ANDed together.
     * <p>
     * This method is not supported in the Syntax Check mode.
	 * <p>
	 * Note: You have to close the iterator after iterating the data using {@link ResultIterator#close()}.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>{@code
	 * def filters = [
	 *      Filter.equal("attribute2", "Active")
	 * ]
	 *
	 * def products = api.stream("P", "-attribute1", ["attribute2"], *filters)
	 *
	 * products.each { product ->
	 *      ...
	 * }
	 * }</pre>
	 * </blockquote>
	 *
	 * @param typeCode Type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}.
	 * @param sortBy   String of fields/attributes, separated by a comma, based on which the result set should be sorted. Any field can be prefixed with '-' for descending sort order.
	 * @param fields   List of attributes you want to return in the result, i.e. skip all attributes not contained in the list that the search would return otherwise.
	 * @param filters  Optional list of filters (that will be evaluated using the AND operator).
	 * @return Result implements the Iterator interface, allowing for iteration over all records in the scope of the query.
	 * @see #find(String, int, int, String, List, Filter...)
	 * @see #count(String, Filter...)
	 */
	ResultIterator stream(String typeCode, String sortBy, List<String> fields, Filter... filters);

	/**
	 * Searches for all types of objects with a freestyle query and iterates over the result set.
	 * The filters are ANDed together.
     * <p>
     * This method is not supported in the Syntax Check mode.
	 * <p>
	 * Note: You have to close the iterator after iterating the data using {@link ResultIterator#close()}.
	 *
	 * @param typeCode           Type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}.
	 * @param sortBy             String of fields/attributes, separated by a comma, based on which the result set should be sorted. Any field can be prefixed with '-' for descending sort order.
	 * @param fields             List of attributes you want to return in the result, i.e. skip all attributes not contained in the list that the search would return otherwise.
	 * @param distinctValuesOnly Fetches distinct/all values. The use case here is fetching distinct single column values.
	 * @param filters            Optional list of filters (that will be evaluated using the AND operator).
	 * @return Result implements the Iterator interface, allowing for iteration over all records in scope of the query.
	 * @see #find(String, int, int, String, List, boolean, Filter...)
	 * @see #count(String, Filter...)
	 */
	ResultIterator stream(String typeCode, String sortBy, List<String> fields, boolean distinctValuesOnly, Filter... filters);

	/**
	 * Searches for all types of objects with a freestyle query and iterates over the result set.
	 * The filters are ANDed together.
     * <p>
     * This method is not supported in the Syntax Check mode.
	 * <p>
	 * Note: You have to close the iterator after iterating the data using {@link ResultIterator#close()}.
	 *
	 * @param typeCode           Type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}.
	 * @param sortBy             String of fields/attributes, separated by a comma, based on which the result set should be sorted. Any field can be prefixed with '-' for descending sort order.
	 * @param fields             Map of pairs [fieldname : aggregation] holding fields to be returned in the result  where the aggregation can be one of these SUM, AVG, MIN or MAX.
	 * 							 The key of the map is the field name. The value can be a field operator like SUM, AVG, MIN or MAX.
	 * @param distinctValuesOnly Fetches distinct/all values. The use case here is fetching distinct single column values.
	 * @param filters            Optional list of filters (that will be evaluated using the AND operator).
	 * @return Result implements the Iterator interface, allowing for iteration over all records in scope of the query.
	 * @see #find(String, int, int, String, Map, boolean, Filter...)
	 * @see #count(String, Filter...)
	 */
	ResultIterator stream(String typeCode, String sortBy, Map<String,String> fields, boolean distinctValuesOnly, Filter ... filters);

	/**
	 * When given a list of entities (result of api.find,...), it converts the list of meta
	 * attributed entities to a list of maps where each map is a serialized entity whose keys are using field names
	 * from the attribute meta object.
	 * I.e. it converts attributeX fields to names defined in the metadata.
	 *
	 * @param objects List of entities (results of api.find,...).
	 * @return List of serialized entities where keys have the names from metadata.
	 */
	List<Map<String, Object>> namedEntities(List<Object> objects);

	/**
	 * Emits an object to be consumed by an external thread. Requires a blocking emitQueue to be set in the FormulaEngineContext and
	 * a consuming thread to take the emitted objects from the queue. See also FeederFormulaExecutor.
     * <p>
     * This method is not supported in the Syntax Check mode.
	 *
	 * @param obj Object to be emitted/inserted in the emit queue.
	 * @return True if the object is successfully inserted in the queue, false otherwise.
	 * @throws InterruptedException If interrupted while waiting.
	 */
	boolean emit(Object obj) throws InterruptedException;

	/**
	 * Emits the partitioned object - to be consumed by a PA allocation task (a PA calculation task with a feeder formula calling
	 * this method, and a regular formula accepting each emitted partitioned object as its currentItem context, and distributing one or more partitioned object accruals
	 * over the set of PA rows which have contributed to those accruals in some way (by revenue, volume...).
	 *
	 * @param typedId	typedId of the PartitionedObject to emit.
	 * @param slice		{@link DMDataSlice} defines DM slice to which given partitioned object contributes. Make sure it contains valid dateFieldName via {@link DMDataSlice#setDateFieldName(String)}.
	 * @return	True if the object is successfully inserted in the queue, false otherwise.
	 * @throws InterruptedException can throw InterruptedException
	 */
	public boolean emitDMScopedObject(final String typedId, final Object slice) throws InterruptedException;

	/**
	 * Emits the rebate records matching the given search criteria - to be consumed by a PA allocation task (a PA calculation task with a feeder formula calling
	 * this method, and a regular formula accepting each emitted RebateRecord as its currentItem context, and distributing one or more RebateRecord accruals
	 * over the set of PA rows which have contributed to those accruals in some way (by revenue, volume...).
	 *
	 * @param rebateRecordSetLabel	Label of the RebateRecordSet from which to emit records.
	 * @param dateFieldName	Name of the field to be substituted in the RebateRecord calculationBase (a DMDataSlice object that includes a generic time dimension filter).
	 * @param sortBy		RebateRecord field on which to sort.
	 * @param filters		Filters to apply to the RebateRecord search.
	 * @throws InterruptedException if interrupted while waiting
	 */
	void emitRebateRecords(String rebateRecordSetLabel, String dateFieldName, String sortBy, Filter... filters) throws InterruptedException;

	/**
	 * Emits the rebate records matching the given search criteria - to be consumed by a PA allocation task (a PA calculation task with a feeder formula calling
	 * this method, and a regular formula accepting each emitted RebateRecord as its currentItem context, and distributing one or more RebateRecord accruals
	 * over the set of PA rows which have contributed to those accruals in some way (by revenue, volume...).
	 *
	 * @param dateFieldName	Name of the field to be substituted in the RebateRecord calculationBase (a DMDataSlice object that includes a generic time dimension filter).
	 * @param sortBy		RebateRecord field on which to sort.
	 * @param filters		Filters to apply to the RebateRecord search.
	 * @throws InterruptedException if interrupted while waiting
	 */
	void emitRebateRecords(String dateFieldName, String sortBy, Filter... filters) throws InterruptedException;

	/**
	 * Finds a PersistedObject with the given typedId and emits it if it exists. The assumption is that there is a receiving consumer, as in {@link #emitRebateRecords}.
	 * @param typedId	typedId of the PersistedObject to emit.
	 * @return False if failed.
	 * @throws InterruptedException If interrupted while waiting.
	 */
	boolean emitPersistedObject(String typedId) throws InterruptedException;

	/**
	 * Finds a PersistedObject for the given type and ID and emits it if it exists. The assumption is that there is a receiving consumer, as in {@link #emitRebateRecords}.
	 * @param typeCode	typeCode of the PersistedObject to emit.
	 * @param id		ID of the PersistedObject to emit.
	 * @return False if failed.
	 * @throws InterruptedException If interrupted while waiting.
	 */
	boolean emitPersistedObject(String typeCode, Long id) throws InterruptedException;

	/**
	 * Returns the number of records that a corresponding {@link #find(String, int, int, String, List, Filter...)} would return.
	 *
	 * @param typeCode Type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}.
	 * @param filters  List of {@link Filter} objects that are ANDed together.
	 * @return Number of matching records.
	 */
	int count(String typeCode, Filter... filters);

	/**
	 * Finds a price parameter table object (LT):
	 * <ul>
	 *     <li>of the given name;</li>
	 *     <li>which is valid at the current target date;</li>
	 *     <li>whose Status is Active or Simulation Only (Simulation Only tables will be used only when the logic is run as simulation);</li>
	 *     <li>if executed during simulation, it must have a matching Simulation Set.</li>
	 * </ul>
	 * <p>Performance consideration: There is no caching of the values, so you should consider caching it yourself.</p>
	 * <p>Example:
	 * <pre>
	 * def table = <strong>api.findLookupTable</strong>("PricelistApprovalLevels")
	 *
	 * def row = api.find("MLTV", 0, 1, null,
	 *
	 *                      Filter.equal(<strong>"lookupTable.id", table?.id</strong>),
	 *
	 *                      Filter.equal("name", api.product("BusinessUnit"))
	 *           )?.getAt(0)
	 * </pre>
	 *
	 * @param tableName Name of the price parameter table.
	 * @return Price parameter table object (as Map) or {@code null} if the table does not exists or its validAfter &gt; targetDate.
	 */
	LookupTable findLookupTable(String tableName);

	/**
	 * Returns a filtered list of rows of a price parameter table. It will use a Price Parameter table:
	 * <ul>
	 *     <li>of the given name;</li>
	 *     <li>which is valid at the current target date; </li>
	 *     <li>whose Status is Active or Simulation Only (Simulation Only tables will be used only when the logic is run as simulation);</li>
	 *     <li>if executed during simulation, it must have a matching Simulation Set.</li>
	 * </ul>
	 *
	 * <p>Performance consideration: There is no caching of the values, so you should consider caching it yourself.</p>
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def salesOrgMap = <strong>api.findLookupTableValues("SalesOrg")</strong>.collectEntries{
	 *                        [(it.name) : it.name]
	 *                   }
	 *
	 * return api.option("SalesOrg", salesOrgMap.keySet() as List, salesOrgMap)
	 * </pre>
	 * </blockquote>
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * api.retainGlobal = true
	 *
	 * // for the first time
	 * if (api.global.salesOrgs == null) {
	 *     // store all values as a Map [name : value] to global cache
	 *     api.global.salesOrgs = <strong>api.findLookupTableValues("SalesOrg")</strong>.collectEntries {
	 *                                           [(it.name): it.value]
	 *                            }
	 * }
	 *
	 * return api.global.salesOrgs
	 * </pre>
	 * </blockquote>
	 *
	 * @param tableName Name of the price parameter table.
	 * @param filters   Optional extra {@link Filter}.
	 * @return Price parameter table values or {@code null}.
	 *
	 * @see #findLookupTableValues(String, String, Filter...)
	 */
	List<Object> findLookupTableValues(String tableName, Filter... filters);

	/**
	 * Returns sorted rows of a Price Parameter table (optionally filtered). It will use a Price Parameter table:
	 * 	 <ul>
	 * 	   <li>of the given name;</li>
	 *     <li>which is valid at the current target date;</li>
	 *     <li>whose Status is Active or Simulation Only (Simulation Only tables will be used only when the logic is run as simulation);</li>
	 *     <li>if executed during simulation, it must have a matching Simulation Set.</li>
	 * </ul>
	 *
	 * Performance consideration: There is no caching of the values, so you should consider caching it yourself.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * api.retainGlobal = true
	 *
	 * // for the first time
	 * if (api.global.salesOrgs == null) {
	 *     // store all values as a Map [name : value] to global cache
	 *     api.global.salesOrgs = <strong>api.findLookupTableValues("SalesOrg", "-name")</strong>.collectEntries {
	 *         [(it.name): it.value]
	 *     }
	 * }
	 *
	 * return api.global.salesOrgs
	 * </pre>
	 * </blockquote>
	 *
	 * @param tableName Name of the price parameter table.
	 * @param sortBy  Name of column to be used for sorting of the rows. If the char "-" precedes the name, it will do a reverse order (for example "-key1"). Use {@code null} if no sorting is required.
	 * @param filters   Optional extra {@link Filter}.
	 * @return Price parameter table values or {@code null}.
	 *
	 * @see #findLookupTableValues(String, Filter...)
	 */
	List<Object> findLookupTableValues(String tableName, String sortBy, Filter... filters);

	/**
	 * Returns sorted rows of a Price Parameter table (optionally filtered). It will use a Price Parameter table:
	 * 	 <ul>
	 * 	   <li>of the given name;</li>
	 *     <li>which is valid at the current target date;</li>
	 *     <li>whose Status is Active or Simulation Only (Simulation Only tables will be used only when the logic is run as simulation);</li>
	 *     <li>if executed during simulation, it must have a matching Simulation Set.</li>
	 * </ul>
	 *
	 * Performance consideration: There is no caching of the values, so you should consider caching it yourself.
	 * It optionally also takes a cacheProvider argument (can be null; disabling that function). This provider is essentially a map (e.g. a api.global or api.local entry)
	 * provided by the logic - which is managing the scope or also can prefill that cache.
	 * Missing entries are added to that cache by the method. The purpose of it is to save repeating tableName =&gt; actual table mappings.
	 * The format of the cache is "tableName" =&gt; Cache  Entry, which is again a map with keys "id" (=table ID) and "valueType" (=type code of table entries)
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * api.retainGlobal = true
	 *
	 * // for the first time
	 * if(api.global.ltCache == null) {
	 *   api.global.put("ltCache", [:])
	 * }
	 *
	 * return api.findLookupTableValues("Rebate", api.global.ltCache, null, null)
	 *
	 * </pre>
	 * </blockquote>
	 * @param tableName tableName Name of the price parameter table.
	 * @param fields List of fields to retrieve. Can be null
	 * @param sortBy  Name of column to be used for sorting of the rows. If the char "-" precedes the name, it will do a reverse order (for example "-key1"). Use {@code null} if no sorting is required.
	 * @param filters   Optional extra {@link Filter}.
	 * @return Price parameter table values or {@code null}.
	 */
	List<Object> findLookupTableValues(String tableName, List<String> fields, String sortBy, Filter ... filters);

	/**
	 * Returns sorted rows of a Price Parameter table (optionally filtered). It will use a Price Parameter table:
	 * 	 <ul>
	 * 	   <li>of the given name;</li>
	 *     <li>which is valid at the current target date;</li>
	 *     <li>whose Status is Active or Simulation Only (Simulation Only tables will be used only when the logic is run as simulation);</li>
	 *     <li>if executed during simulation, it must have a matching Simulation Set.</li>
	 * </ul>
	 *
	 * Performance consideration: There is no caching of the values, so you should consider caching it yourself.
	 * It optionally also takes a cacheProvider argument (can be null; disabling that function). This provider is essentially a map (e.g. a api.global or api.local entry)
	 * provided by the logic - which is managing the scope or also can prefill that cache.
	 * Missing entries are added to that cache by the method. The purpose of it is to save repeating tableName =&gt; actual table mappings.
	 * The format of the cache is "tableName" =&gt; Cache  Entry, which is again a map with keys "id" (=table ID) and "valueType" (=type code of table entries)
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * api.retainGlobal = true
	 *
	 * // for the first time
	 * if(api.global.ltCache == null) {
	 *   api.global.put("ltCache", [:])
	 * }
	 *
	 * return api.findLookupTableValues("Rebate", api.global.ltCache, null, null)
	 *
	 * </pre>
	 * </blockquote>
	 * @param tableName tableName Name of the price parameter table.
	 * @param cacheProvider A map object that serves as the cache. Can be null (=no cache used), empty or prefilled
	 * @param fields List of fields to retrieve. Can be null
	 * @param sortBy  Name of column to be used for sorting of the rows. If the char "-" precedes the name, it will do a reverse order (for example "-key1"). Use {@code null} if no sorting is required.
	 * @param filters   Optional extra {@link Filter}.
	 * @return Price parameter table values or {@code null}.
	 */
	List<Object> findLookupTableValues(String tableName, Map<String,Object> cacheProvider, List<String> fields, String sortBy, Filter ... filters);
	/**
	 * Finds active workflows for the given approvable. You can get the same information as displayed in the workflow UI.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def quoteUniqueName = "P-123"
	 * def quote = api.find('Q', 0, 1, Filter.equal("uniqueName", quoteUniqueName)).getAt(0)
	 * return quote ? api.findWorkflowInfo('Q', quote.id) : null
	 * </pre>
	 * </blockquote>
	 *
	 * @param approvableTypeCode Type code of approvable. Approvables are e.g. quotes, rebate agreements, data change requests etc.
	 * @param approvableId       ID of the approvable.
	 * @return Table like structure resembling workflow UI or {@code null} if not found.
	 * @see #findPreviousWorkflowInfo(String, String...)
	 * @see #findPreviousApprovableState(String, String...)
	 */
	WorkflowInfo findWorkflowInfo(String approvableTypeCode, Object approvableId);

	/**
	 * Finds the previous approvable entity state if some DENIED or WITHDRAWN workflow for the approvable exists.
	 * @param typedId Approvable entity typed ID.
	 * @param workflowStatuses Only DENIED or WITHDRAWN are supported.
	 * @return State of the approvable entity when its previous workflow was DENIED or WITHDRAWN.
	 * @see #findWorkflowInfo(String, Object)
	 */
	Map<String, Object> findPreviousApprovableState(String typedId, String... workflowStatuses);

	/**
	 * Finds the approvable entity's previous workflow history if some DENIED or WITHDRAWN workflow for the approvable exists.
	 * @param typedId Approvable entity typed ID.
	 * @param workflowStatuses Only DENIED or WITHDRAWN are supported.
	 * @return Workflow history, see net.pricefx.workflowengine.WorkflowInfo.
	 * @see #findWorkflowInfo(String, Object)
	 */
	WorkflowInfo findPreviousWorkflowInfo(String typedId, String... workflowStatuses);

	/**
	 * Finds all assignment records for the given customerId.
	 * Also matched by CustomerGroup definitions.
	 *
	 * @param customerId Customer whose assignments are searched for.
	 * @return List of assignment records.
	 */
	List<CustomerAssignment> findCustomerAssignments(String customerId);

	/**
	 * Returns a list of price lists that match the targetDate (i.e. list's targetDate &lt;= targetDate and expiryDate &gt; targetDate).
	 *
	 * @param targetDate targetDate to use.
	 * @param listName   Price list name. If {@code null}, the list name is not included in search.
	 * @return List of matching price lists.
	 */
	List<Pricelist> findPricelists(Date targetDate, String listName);

	/**
	 * Returns a list of simulations that match the targetDate (i.e. list's targetDate &lt;= targetDate).
	 *
	 * @param targetDate     targetDate to use.
	 * @param simulationName Simulation name. If {@code null}, the simulation name is not included in search.
	 * @return List of matching simulations.
	 */
	List<Simulation> findSimulations(Date targetDate, String simulationName);

	/**
	 * Returns a list of price grids that match the targetDate (i.e. price grid's targetDate &lt;= targetDate).
	 *
	 * @param targetDate    targetDate to use.
	 * @param priceGridName Price grid name. If {@code null}, the price grid name is not included in search.
	 * @return List of matching price grids.
	 */
	List<PriceGrid> findPriceGrids(Date targetDate, String priceGridName);

	/**
	 * Returns a list of calculated field sets that match the targetDate.
	 *
	 * @param targetDate Target date for the CFS.
	 * @param cfsName    CFS name. If {@code null}, the CFS name is not included in search.
	 * @return List of matching calculated field sets.
	 */
	List<CalculatedFieldSet> findCalculatedFieldSets(Date targetDate, String cfsName);

	/**
	 * Returns a list of manual price lists that match the targetDate (i.e. list's validAfter &lt;= targetDate and expiryDate &gt; targetDate).
	 *
	 * @param targetDate targetDate to use.
	 * @param listName   Price list name. If {@code null}, the list name is not included in search.
	 * @return List of matching price lists.
	 */
	List<ManualPricelist> findManualPricelists(Date targetDate, String listName);

	/**
	 * Returns a list of summary items. One item is returned for every matching (= same SKU) PLI or XPLI which is in an approved list.
	 *
	 * @param sku SKU to filter on.
	 * @return List of summary items.
	 */
	List<SummaryCalculableObjectItem> findApprovedPricelistItems(String sku);

	/**
	 * Retrieves a list of customers which match the given CustomerGroup (the first 200 entries are returned).
	 *
	 * @param group Group to match.
	 * @return List of customers.
	 */
	List<Customer> findCustomersInGroup(CustomerGroup group);

	/**
	 * Retrieves a list of customers which match the given CustomerGroup with custom paging and sorting.
	 *
	 * @param group Customer group.
	 * @param startRow Starting row number (maximum 200 rows returned at once).
	 * @param sortBy   Sort key. {@code null} if no sorting is required.
	 * @return List of customers.
	 */
	List<Customer> findCustomersInGroup(CustomerGroup group, int startRow, String sortBy);

	/**
	 * Returns a user entered value that overrides the calculated result of a logic element.
     * <p>
     * Note: The manualOverride works as a flag - when it is set for an element, the final
     * value of that element is taken from the manually entered value and the calculation of that
     * element does not run at all. If this flag is not set, the final value is the calculated result.
	 *
	 * @param elementName Formula/result element that is overridden.
	 * @return Overridden value or {@code null} if not overridden.
	 */
	Object getManualOverride(String elementName);

	/**
	 * Removes a user entered value that overrides the calculated result of a logic element.
     * <p>
     * If you have a condition to remove the manual override, it is recommended to put this condition
     * in a separate logic element which is always executed (not in the same one which has the manual
     * override allowed; the reason is that once the user enters a value manually, the element calculation
     * does not run at all, so the condition would not be evaluated).
	 * <p>
     * Note also that a logic element with api.removeManualOverride() must
     * be before an element which is cleared.
     *
	 * @param elementName Formula/result element that is overridden.
	 */
	void removeManualOverride(String elementName);

	/**
	 * Aborts the current item (e.g. price list line, quote line) calculation - i.e. stops the execution of subsequent elements of the logic.
	 * The subsequent logic elements are not executed and
	 * the calculation engine continues with calculation of the next item.
	 * <p><strong>Note</strong>: This function only sets the abortCalculation flag, so the element where api.abortCalculation() is called finishes its execution and
	 * still returns a value (but it is the last element returning a value).
	 * <p>This function is often called together with {@link #isSyntaxCheck()} after the parameters are gathered.
	 *
	 * <p>Example:
	 * <blockquote>
	 * <pre>
	 * if (api.isSyntaxCheck()) {
	 *   <strong>api.abortCalculation()</strong>
	 * }
	 * </pre>
	 * </blockquote>

	 * <p>Example:
	 * <blockquote>
	 * <pre>
	 * def price = ...
	 *
	 * if (!price) {
	 *   <strong>api.abortCalculation()</strong>
	 * }
	 *
	 * return price  //this line will still be executed during the Syntax Check, the logic will be aborted only after this element finishes
	 * </pre>
	 * </blockquote>
	 */
	void abortCalculation();

	/**
	 * Aborts the current item calculation. The subsequent logic elements are not executed and
	 * the calculation engine continues with calculation of the next item.
	 * This function only sets the abortCalculation flag, so the element where abortCalculation is called
	 * still returns a value (and is the last element returning a value).
	 * This function can be called together with {@link #isSyntaxCheck()} after the parameters are gathered.
	 *
	 * <p>
	 * @param keepOld Determines if the old outputs should be retained. Only applicable for quotes and contracts calculation.
	 */
	void abortCalculation(boolean keepOld);

	/**
	 * Aborts the current item parameter determination. This helps in a logic with many elements to tell the engine
	 * not to spend more time on the next elements when the logic author is sure that there is no other statement
	 * that would generate any new inputs.
	 * <p>
	 * This function can be called together with {@link #isSyntaxCheck()} after the parameters are gathered.
	 *
	 */
	void abortSyntaxCheck();

	/**
	 * Throws a calculation exception with the specified message and aborts the calculation logic execution immediately.
	 * <p>Sample use cases:
	 * <ul>
	 *     <li>In workflow logics - When the workflow logic cannot find important/vital data needed for building
	 *         the approval steps, it can throw an exception which immediately stops the submission process
	 *        and the Quote will stay in the Draft state.</li>
	 *     <li>In Quote logics - To stop processing the Quote line calculation when the important inputs (or values) are not available.
	 *         If the calculation was called as part of the submission process, the process stops and the Quote stays in the Draft state.</li>
	 * </ul>
	 *
	 * @param errorMessage Error message that is displayed.
	 * @throws XExpression Custom calculation exception.
	 */
	void throwException(String errorMessage) throws XExpression;

	/**
	 * Sets the dirty flag on the specified item, i.e. the second calculation run will be triggered and this item will be included in the 2nd pass.
	 * If used in a MatrixPricelist/MatrixSimulation (i.e. on objects that have a secondary key), all items with that SKU are marked as dirty.
	 * <p>
	 * See also {@link #getIterationNumber()}, {@link #markItemClean()}.
	 * <p>
	 * The maximum number of dirty passes is defined by the server configuration constant {@code maxDirtyCalculations} defined for the whole instance.
	 * The default is one dirty pass.
	 *
	 * @param sku Item to have the dirty flag set.
	 * @see #markItemClean(String)
	 */
	void markItemDirty(String sku);

	/**
	 * Sets the dirty flag on the specified item, i.e. the second calculation run will be triggered and this item will be included in the 2nd pass.
	 * The maximum number of dirty passes is defined by the server configuration constant {@code maxDirtyCalculations} defined for the whole instance.
	 * The default is one dirty pass.
	 * <p>
	 * See also {@link #getIterationNumber()}, {@link #markItemClean()}.
	 *
	 * @param sku Item to have the dirty flag set.
	 * @param key2 key2 field.
	 * @see #markItemClean(String, String)
	 */
	void markItemDirty(String sku, String key2);

	/**
	 * Sets the dirty flag on the specified item, i.e. the second calculation run will be triggered and this item will be included in the 2nd pass.
	 * The maximum number of dirty passes is defined by the server configuration constant {@code maxDirtyCalculations} defined for the whole instance.
	 * The default is one dirty pass.
	 * <p>
	 * See also {@link #getIterationNumber()}, {@link #markItemClean()}.
	 *
	 * @see #markItemClean()
	 */
	void markItemDirty();

	/**
	 * Removes the dirty flag from the current item set previously by {@link #markItemDirty()}.
	 *
	 * @see #markItemDirty()
	 */
	void markItemClean();

	/**
	 * Removes the dirty flag from the specified matrix item set previously by {@link #markItemDirty(String, String)}.
	 *
	 * @param sku Item from which to remove the dirty flag.
	 * @param key2 key2 field.
	 * @see #markItemDirty(String, String)
	 */
	void markItemClean(String sku, String key2);

	/**
	 * Removes the dirty flag from the specified item set previously by {@link #markItemDirty(String)}.
	 *
	 * @param sku Item from which to remove the dirty flag.
	 * @see #markItemDirty(String)
	 */
	void markItemClean(String sku);

	/**
	 * Determines if the calculation is a dirty pass. It does not necessarily mean it is a 2nd pass calculation.
	 *
	 * @return True if is in a dirty pass.
	 * @deprecated Use {@link #getIterationNumber()} &gt; 0 instead.
	 */
	boolean isSecondPass();

	/**
	 * Returns the number of iterations in which this list has been calculated during this cycle.
	 * <ul>
	 *   <li>0 means: initial calculation (non-dirty)</li>
	 *   <li>1 means: 1st dirty pass, {@link #markItemDirty()} was called in the previous pass</li>
	 *   <li>2 means: 2nd dirty pass, {@link #markItemDirty()} was called in the previous pass</li>
	 *   <li>...and so on.</li>
	 * </ul>
	 * The maximum number of dirty passes is defined by the server configuration constant {@code maxDirtyCalculations} defined for the whole instance.
	 * The default is one dirty pass.
	 * @return Iteration number
	 */
	int getIterationNumber();

	/**
	 * Formats a string like {@link java.lang.String#format}.
	 *
	 * @param locale Locale (as a String). Use only simple locale names like "de" or "fr"
	 * @param format Format string.
	 * @param args   Arguments to pass on.
	 * @return Formatted string
	 */
	String formatString(String locale, String format, Object... args);

	/**
	 * Formats a number as a String. The same semantics as in {@link java.text.NumberFormat} and {@link java.text.DecimalFormat#applyPattern(String)} apply.
	 *
	 * @param pattern Pattern string.
	 * @param n       Number to format.
	 * @return Formatted number as a String.
	 */
	String formatNumber(String pattern, Object n);

	/**
	 * Formats a number as a String. The same semantics as in {@link java.text.NumberFormat} and {@link java.text.DecimalFormat#applyPattern(String)}  apply.
	 *
	 * @param pattern Pattern string.
	 * @param n       Number to format.
	 * @param locale  Locale to use for formatting. Use only simple locale names like "de" or "fr"
	 * @return Formatted number as a String.
	 */
	String formatNumber(String locale, String pattern, Object n);

	/**
	 * A utility method to decode a JSON string into a Map representation.
	 *
	 * @param json Input JSON string.
	 * @return Parsed JSON as a Map.
	 */
	Object jsonDecode(String json);

	/**
	 * A utility method to decode a JSON string into a List representation.
	 *
	 * @param json Input JSON string.
	 * @return Parsed JSON as a List.
	 */
	Object jsonDecodeList(String json);

	/**
	 * A utility method to encode an object to a JSON string. Omits all null values in the output.
	 *
	 * @param input Input object, usually a map.
	 * @return Encoded object
	 */
	Object jsonEncode(Object input);

	/**
	 *
	 * A utility method to encode an object to a JSON string.
	 *
	 * @param input Input object, usually a map.
	 * @param omitNull Determines if null values should be omitted.
	 * @return Encoded object
	 */
	Object jsonEncode(Object input, boolean omitNull);

	/**
	 * Marks the current logic element as a critical alert and provides the given message.
	 * <ul>
	 *   <li>In a price grid, price list or in a quote line item, it marks the whole record with a dark red background color.</li>
	 *   <li>In the calculation results (e.g. in Quote line results), it displays a red circle with an exclamation mark next to the element.</li>
	 * </ul>
	 * <p>To display the error message for price grids and price lists, click the arrow in the Detail column.
	 * <p>It is displayed only if called from an element that has its 'Display mode' enabled for the given context ('PriceGrids', 'Pricelists' or 'Quote Configurator').
	 * <p><strong>Note</strong>: When used in a quote line context, it blocks submission of the quote.
	 *
	 * @param msg Alert message.
	 * @see #redAlert(String)
	 * @see #yellowAlert(String)
	 */
	void criticalAlert(String msg);

	/**
	 * Marks the current logic element with a red alert and provides the given message.
	 * <ul>
	 *   <li>In a price grid, price list or in a quote line item, it marks the corresponding field background with a red background color.</li>
	 *   <li>In the calculation results (e.g. in Quote line results), it displays a plain red circle next to the element.</li>
	 * </ul>
	 * <p>To display the error message for price grids and price lists, click the arrow in the Detail column.
	 * <p>It is displayed only if called from an element that has its 'Display mode' enabled for the given context ('PriceGrids', 'Pricelists' or 'Quote Configurator').
	 *
     * @param msg Alert message.
	 * @see #criticalAlert(String)
	 * @see #yellowAlert(String)
	 */
	void redAlert(String msg);

	/**
	 * Marks the current logic element with a yellow alert and provides the given message.
	 * <ul>
	 *   <li>In a price grid, price list or in a quote line item, it marks the corresponding field background with a yellow background color.</li>
	 *   <li>In the calculation results (e.g. in Quote line results), it displays a plain yellow circle next to the element.</li>
	 * </ul>
	 * <p>To display the error message for price grids and price lists, click the arrow in the Detail column.
	 * <p>The Alert is displayed only if called from an element that has its 'Display mode' enabled for the given context ('PriceGrids', 'Pricelists' or 'Quote Configurator').
	 *
	 * @param msg Alert message.
	 * @see #criticalAlert(String)
	 * @see #redAlert(String)
	 */
	void yellowAlert(String msg);

	/**
	 * Marks the current formula element with the given alert message.
	 * If one was already set by means of a yellowAlert, redAlert or criticalAlert call, then that message is replaced.
	 * When not raising an alert through any of these calls, but still setting a message with this method,
	 * the message will be displayed wherever the calculation results are shown with alert messages, but without any yellow or red background.
	 *
	 * @param msg Alert message.
	 */
	void setAlertMessage(String msg);

	/**
	 * @deprecated
	 * Calls {@link #buildFlexChart(String, Object) buildFlexChart(null, definition)}.
	 * See the full documentation in {@link #buildFlexChart(Object)}.
	 *
	 * @param definition the definition of the {@code FlexChart}.
	 * @return ResultFlexChart object.
	 * @see #buildHighchart(Map)
	 * @see #buildFlexChart(String, Object)
	 */
	ResultFlexChart buildFlexChart(Object definition);

	/**
	 * @deprecated
	 * Creates a {@link FlexChart} from the passed {@code definition} by
	 * merging it with {@code options} defined within the
	 * {@code baseTemplateToUse} template.
	 * <p>
	 * Although it may change in future releases, for now, a
	 * {@code FlexChart} is nothing else than a {@code Highcharts}
	 * chart. Therefore, the passed {@code definition} (also applies to the
	 * {@code baseTemplateToUse} template) must denote
	 * {@code Highcharts options} - please consult {@code Highcharts}
	 * documentation for more details, and <b>note that FlexCharts must comply
	 * to the outdated Highcharts release 4</b>.
	 * <p>
	 * Since {@code Highcharts options} are normally specified by a
	 * {@code JavaScript object literal}, the passed
	 * {@code definition} must be either a {@code JSON} string
	 * representing such {@code object literal}, or any {@code object}
	 * which can be mapped to such {@code JSON} representation. Therefore,
	 * typically the {@code definiton} will be passed as a
	 * {@code map of maps}.
	 * <p>Example:
	 * <blockquote>
	 * <pre>
	 * def definition = [
	 *   chart: [
	 *     type: "bar"
	 *   ],
	 *   title: [
	 *     text: "A common FlexChart"
	 *   ],
	 *   series: [
	 *    [
	 *      data: [10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21]
	 *    ]
	 *   ]
	 * ]
	 *
	 * api.buildFlexChart(definition)
	 * </pre>
	 * </blockquote>
	 * Note that not all {@code Highcharts options} are supported.
	 * Moreover, usage of some options is prohibited deliberately. These are
	 * currently {@code global}, {@code events} and every
	 * {@code JavaScript function}-based options.
	 * <p>
	 * Note: {@code FlexChart} templates are defined by an administrator in
	 * the '{@code Configuration}' section.
	 * <p>
	 *
	 * @param baseTemplateToUse Name of the template to use as a basis for merging with
	 *                          the passed {@code definition}; if {@code null}, the
	 *                          predefined Default template is used.
	 * @param definition        Definition of the {@code FlexChart}; currently, this
	 *                          will typically be a string or map of maps containing
	 *                          {@code Highcharts options}.
	 * @return Helper object linking the resulting {@code FlexChart},
	 * {@code baseTemplateToUse} and type of the result - which
	 * will always be 'FLEXCHART' here.
	 * @see <a href="http://api.highcharts.com/highcharts">http://api.highcharts.com/highcharts</a>
	 * @see <a href="http://www.highcharts.com/docs">http://www.highcharts.com/docs</a>
	 * @see <a href="http://json.org/">http://json.org</a>
	 * @see #buildFlexChart(Object)
	 * @see #buildHighchart(Map)
	 */
	ResultFlexChart buildFlexChart(String baseTemplateToUse,
	                               Object definition);

	/**
	 * Creates a {@link ResultHighchart} from the passed {@code definition}.
	 * <br>
	 * As opposed to {@code FlexChart}, the definition is passed straight to the
	 * {@code Highcharts} library, currently release 8.2. There is no validation,
	 * no templates merging, and JavaScript functions are not supported.
	 * @param definition        Definition of the {@code Highchart} as a map of maps and arrays.
	 * @return ResultHighchart object
	 * @see <a href="https://www.highcharts.com/docs">https://www.highcharts.com/docs</a>
	 * @see <a href="https://api.highcharts.com/highcharts">https://api.highcharts.com/highcharts</a>
	 * @see #buildFlexChart(Object)
	 */
	ResultHighchart buildHighchart(final Map<String, ?> definition);

	/**
	 * Creates a {@link ResultHighmap} from the passed {@code definition}.
	 * <br>
	 * As opposed to {@code FlexChart}, the definition is passed straight to the
	 * {@code Highmaps} library, currently release 8.2. There is no validation,
	 * no templates merging, and JavaScript functions are not supported.
	 * {@code HighChart} provide better performance than {@code FlexChart}.
	 * @param definition        Definition of the {@code Highmap} as a map of maps and arrays.
	 * @return buildHighmap object
	 * @see <a href="https://www.highcharts.com/docs">https://www.highcharts.com/docs</a>
	 * @see <a href="https://api.highcharts.com/highmaps">https://api.highcharts.com/highmaps</a>
	 */
	ResultHighmap buildHighmap(final Map<String, ?> definition);

	/**
	 * Creates a histogram that adapts to an unknown data distribution.
	 * It keeps a more or less constant resolution throughout the data range by increasing
	 * the resolution where the data is more dense. For example, if the data has
	 * such a distribution that most of the values lie in the 0-5 range and only a few are
	 * in the 5-10 range, the histogram would adapt and assign more counting buckets to
	 * the 0-5 range and less to the 5-10 range.
	 * <p>
	 * The histogram provides a method to obtain the accumulative density function
	 * for a given data point (getValueForPercentile), and a method to obtain the data point that splits the
	 * data set at a given percentile (getValueForPercentile).
	 * @return New AdaptiveHistogram object
	 */
	Object newAdaptiveHistogram();

    /**
     * Creates a representation of a System portlet, like the Quotes list, To Do list, Quick Product Search, etc.
     * that are typically display on the application's home screen.
     * <br/>
     * Rendering these portlets is currently supported only in Dashboards.
     * For a full list of the available portlets, see {@link SystemPortlet}.
     * <br/>
     * Note: This is available in Unity UI only. No backporting planned for the Classic UI.
     *
     * @param type  the desired portlet type (constants for available types are defined in {@link SystemPortlet})
     * @return New {@link SystemPortlet} object
     * @since 7.0
     */
	Object newSystemPortlet(final String type);

	/**
	 * Creates a summary query object that can be parameterized and is then subsequently used in {@link #runSummaryQuery(ItemSummaryQuery)}.
	 * This query runs against price lists.
	 *
	 * @return Query definition object
	 */
	ItemSummaryQuery<Pricelist> getPricelistSummaryQuery();

	/**
	 * Creates a summary query object that can be parameterized and is then subsequently used in {@link #runSummaryQuery(ItemSummaryQuery)}.
	 * This query runs against simulations.
	 *
	 * @return Query definition object
	 */
	ItemSummaryQuery<Simulation> getSimulationSummaryQuery();

	/**
	 * Creates a summary query object that can be parameterized and is then subsequently used in {@link #runSummaryQuery(ItemSummaryQuery)}.
	 * This query runs against price grids.
	 *
	 * @return Query definition object
	 */
	ItemSummaryQuery<PriceGrid> getPriceGridSummaryQuery();

	/**
	 * Creates a summary query object that can be parameterized and is then subsequently used in {@link #runSummaryQuery(ItemSummaryQuery)}.
	 * This query runs against rebate records.
	 *
	 * @return Query definition object
	 */
	ItemSummaryQuery<RebateRecordSet> getRebateRecordSummaryQuery();

	/**
	 * Runs (executes) a summary query defined by the parameterized query object which was created by one of the getXXXSummaryQuery calls.
	 *
	 * @param query Query to execute.
	 * @return Result set of the query
	 */
	List<Map<String, Object>> runSummaryQuery(ItemSummaryQuery<?> query);

	/**
	 * Creates a configurator entry object and adds the specified input parameter as the first parameter to the configurator entry.
	 * If the parameter had a value in the previous round trip, that value is set automatically.
	 *
	 * @param type Input type of the parameter. Valid enum values are:
	 * 	PRODUCT,	PRODUCTGROUP,	CUSTOMER,	CUSTOMERGROUP,	USERENTRY,	STRINGUSERENTRY, OPTION,
	 * 	INTEGERUSERENTRY,	BOOLEANUSERENTRY,	DATEUSERENTRY,	TIMEUSERENTRY,	DATETIMEUSERENTRY,	MULTITIERENTRY,
	 * 	INPUTMATRIX, OPTIONS,	MATRIXLOOKUP,	LOOKUP,	FILTERBUILDER,	DMFILTERBUILDER,	DMFIELD,	DMDIMFILTER,
	 * 	CONFIGURATOR,	HIDDEN.
	 * @param name Name of the parameter.
	 * @return Configurator entry with the input parameter added.
	 */
	ConfiguratorEntry createConfiguratorEntry(InputType type, String name);

	/**
	 * Creates a configurator entry object and adds the specified input parameter as the first parameter to the configurator entry.
	 * If the parameter had a value in the previous round trip, that value is set automatically.
	 *
	 * @param type Input type of the parameter. Valid enum values are:
	 * 	PRODUCT,	PRODUCTGROUP,	CUSTOMER,	CUSTOMERGROUP,	USERENTRY,	STRINGUSERENTRY, OPTION,
	 * 	INTEGERUSERENTRY,	BOOLEANUSERENTRY,	DATEUSERENTRY,	TIMEUSERENTRY,	DATETIMEUSERENTRY,	MULTITIERENTRY,
	 * 	INPUTMATRIX, OPTIONS,	MATRIXLOOKUP,	LOOKUP,	FILTERBUILDER,	DMFILTERBUILDER,	DMFIELD,	DMDIMFILTER,
	 * 	CONFIGURATOR,	HIDDEN.
	 * @param name       Name of the parameter.
	 * @param targetDate Target date for the input parameter (not applicable to all input types).
	 * @return Configurator entry with the input parameter added.
	 */
	ConfiguratorEntry createConfiguratorEntry(InputType type, String name, Date targetDate);

	/**
	 * Creates an empty configurator entry object. Must be filled with input parameters by calling createParameter on the object.
	 *
	 * @return Configurator entry
	 */
	ConfiguratorEntry createConfiguratorEntry();

	/**
	 * Creates a configurator entry array object which is to be returned from
	 * a logic element. Allows dynamic setup of the configurator entries.
	 *
	 * @param entries
	 *            0..N configurator entries. Supports also list nesting.
	 * @return Configurator entry array object with defined methods for
	 *         further anrichment of the array as
	 *         {@code ConfiguratorEntryArray#addEntry(ConfiguratorEntry)},
	 *         {@code ConfiguratorEntryArray#setEntries(List)} or
	 *         {@code ConfiguratorEntryArray#getEntries()}.
	 */
	ConfiguratorEntryArray createConfiguratorEntryArray(Object ... entries);

	/**
	 *
	 *
	 * @param configuratorName Name of the configurator, similar to other options' names.
	 * @param formulaName      Name of the configurator formula which is used to drive the input selections.
	 * @return Result of the configurator
	 */
	Object configurator(String configuratorName, String formulaName);

	/**
	 *
	 *
	 * @param configuratorName Name of the configurator, similar to other options' names.
	 * @param formulaName      Name of the configurator formula which is used to drive the input selections.
	 * @param resultName	   Result that should be returned (instead of the full configurator results map).
	 * @return Result of the configurator
	 */
	Object configurator(String configuratorName, String formulaName, Object resultName);

	/**
	 *
	 *
	 * @param configuratorName Name of the configurator, similar to other options' names.
	 * @param formulaName      Name of the configurator formula which is used to drive the input selections.
	 * @param width			   Width of the configurator window. Can be a number (in pixels) or a percentage string.
	 * @param height		   Height of the configurator window. Can be a number (in pixels) or a percentage string.
	 * @return Result of the configurator
	 */
	Object configurator(String configuratorName, String formulaName,Object width, Object height);

	/**
	 *
	 *
	 * @param configuratorName Name of the configurator, similar to other options' names.
	 * @param formulaName      Name of the configurator formula which is used to drive the input selections.
	 * @return Result of the configurator
	 */
	Object inlineConfigurator(String configuratorName, String formulaName);

	/**
	 * Parses a date (entered as a String) according to the given pattern.
	 * <p>
	 * Hint: For parsing datetimes (i.e. dates also including information about the time of the day) use the {@link #parseDateTime(String, String)} method instead.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def date = api.parseDate("yyyy-MM-dd", "2019-01-23")
	 * </pre>
	 * </blockquote>
	 *
	 * @param pattern Pattern for parsing (see <a href="http://www.joda.org/joda-time/apidocs/index.html?org/joda/time/format/DateTimeFormat.html">DateTimeFormat</a> for the pattern format specification).
	 * @param date    Date as a String in the format specified by the {@code pattern}.
	 * @return Date
	 * @see #parseDateTime(String, String)
	 * @see #parseDateTime(String, String, int, int)
	 */
	Date parseDate(String pattern, String date);

	/**
	 * Parses a datetime (entered as a String) according to the given pattern.
	 * <p>
	 * Note: This method will use the UTC time zone for parsing (unless the provided {@code pattern} uses a syntax explicitly
	 * specifying a time zone). If a different time zone is required, use the {@link #parseDateTime(String, String, int, int)} method which allows to
	 * set a time zone by providing an offset to UTC in hours and minutes.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def now = api.parseDateTime("yyyy-MM-dd'T'HH:mm:ss", "2019-01-23T13:44:52")
	 * now.getYear()            // 2019
	 * now.getMonthOfYear()     // 1
	 * now.getDayOfMonth()      // 23
	 * now.getHourOfDay()       // 13
	 * now.getMinuteOfHour()    // 44
	 * now.getSecondOfMinute()  // 52
	 * </pre>
	 * </blockquote>
	 *
	 * @param pattern  Pattern for parsing (see <a href="http://www.joda.org/joda-time/apidocs/index.html?org/joda/time/format/DateTimeFormat.html">DateTimeFormat</a> for the pattern format specification).
	 * @param datetime datetime string to be parsed
	 * @return Datetime
	 * @see #parseDate(String, String)
	 * @see #parseDateTime(String, String, int, int)
	 */
	DateTime parseDateTime(String pattern, String datetime);

	/**
	 * Parses a datetime (entered as a String) according to a given pattern and time zone.
	 * <p>
	 * Note: This method will always use the provided time zone, regardless of the possible time zone set explicitly in {@code datetime}.
	 *
	 * @param pattern         Pattern for parsing (see <a href="http://www.joda.org/joda-time/apidocs/index.html?org/joda/time/format/DateTimeFormat.html">DateTimeFormat</a> for the pattern format specification).
	 * @param datetime        Datetime as a String in the format specified by the {@code pattern}.
	 * @param tzHoursOffset   Offset in hours from UTC, from -23 to +23.
	 * @param tzMinutesOffset Offset in minutes from UTC, from -59 to +59.
	 * @return Datetime
	 * @see #parseDateTime(String, String)
	 * @see #parseDate(String, String)
	 */
	DateTime parseDateTime(String pattern, String datetime, int tzHoursOffset, int tzMinutesOffset);

	/**
	 * Creates a cross-list filter expression based on metadata.
	 * <p>
	 * Example:
	 * You need to filter across multiple price lists on items that have the value "A" in the column with the elementName "XYZ".
	 * Now the elementName XYZ might be mapped to a different attributeXX column for every list.
	 * So a simple attributeXX=A filter does not work.
	 * <p>
	 * This method does a metadata lookup and creates a proper filter clause based on it for every list.
	 * For every list this translates into something like:
	 * (attributeXX = A AND listId = &lt;id&gt;).
	 * These filters are ORed together.
	 *
	 * @param elementName  Element name (comes from the logic element name or MPL integration tag).
	 * @param fieldValue   Value (= operator applied) of that attribute.
	 * @param listTypedIds List of typedIds of the "header" objects. E.g. 34.PL. Supported type codes are PL, MPL, PG and SIM.
     * @return Filter object
	 */
	Filter createElementNameFilter(String elementName, Object fieldValue, String... listTypedIds);

	/**
	 * Gets the current complete calculation results of the given item (e.g., price list item).
	 * <p>
	 * Note: These are not returned when using other methods (e.g., {@link #pricelistItem(String, String)}).
	 *
	 * @param typedId typedId of the item, e.g. 34.PLI. Supported type codes are PLI, XPLI, SIMI, XSIMI, PGI, XPGI and RR.
	 * @return Map of individual calculation results with 'Element Name' as the key.
	 */
	Map<String, Object> getItemCompleteCalculationResults(String typedId);

	/**
	 * Gets the current complete active (i.e. approved) calculation results of the given item.
	 *
	 * @param typedId typedId of the item, e.g. 34.PLI. Supported type codes are PGI and XPGI.
	 * @return Map of individual calculation results with 'Element Name' as the key.
	 */
	Map<String, Object> getItemActiveCalculationResults(String typedId);

	/**
	 * Issues an HTTP call to an external entity. Defaulting to POST request and JSON content type.
     * Arbitrary addresses/URLs are allowed with the exception of internal addresses
     * (in particular xxx.pricefx.net hosts are blacklisted for security reasons).
	 * <p>
     * If called from a logic element, it has the same timeout as defined for that element.
	 *
	 * @param url Full URL to the HTTPS or HTTP endpoint.
	 * @param body Body of the request.
     * @see #httpCall(String, String, String, String, Map[])
     * @return Map containing two elements: 'responseBody' and 'errorCode'. If the content type is JSON and data is returned, responseBody will actually also be a parsed Map and not the raw data.
	 * @throws XExpression If the HTTP call fails.
	 */
	Object httpCall(String url, String body) throws XExpression;

	/**
	 * Triggers a HTTPS request towards the Pricefx server, either local one or an external one.
     * Arbitrary addresses/URLs are allowed with the exception of internal addresses
     * (in particular xxx.pricefx.net hosts are blacklisted for security reasons). If this method is
	 * called from a logic element, it has the same timeout as defined for that element.
	 * <p>
	 * Note: Use this method with caution, it has some drawbacks. For details see
	 * <a href="https://pricefx.atlassian.net/wiki/display/KB/Avoid+api.boundCall">Avoid api.boundCall</a>.
	 * <p>
	 * To be able to use this function, you need to set up a bound partition first in Configuration &gt; All
	 * Modules &gt; Bound Partitions. This bound partition must always be set up against some technical account created
	 * specifically for the bound call purpose. When calling the Pricefx backend, there is a default
	 * size limit of 200k (set in jetty.maxFormContentSize).
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>{@code
	 * def items = ...
	 * def requestBody = []
	 * items.each { item ->
	 *
	 *     def record = [
	 *         "name": "Increase",
	 *         "sku": item.sku,
	 *         "attribute1": item.newValue,
	 *     ]
	 *
	 *     requestBody.add([
	 *         data: record,
	 *         oldValues: [:]
	 *     ])
	 * }
	 * def resp = api.boundCall("localhost", "/integrate/PX/batch", api.jsonEncode(requestBody), true)
	 * }</pre>
	 * </blockquote>
	 *
	 * @param uniqueName            Unique name of the bound partition.
	 * @param relativeUrl           Relative URL to the Pricefx server command. First slash excluded.
	 * @param body                  Body of the request.
	 * @param responseBodyAsAString If the first param is true, the body is not converted to JSON but returned as a String.
	 * @return Map containing two elements: 'responseBody' and 'errorCode'.
	 * @throws XExpression If the bound partition is not set or the call fails.
	 * @see <a href="https://qa.pricefx.eu/pricefx-api/json/index.html">Pricefx backend API</a>
	 */
	Object boundCall(String uniqueName, String relativeUrl, String body, Boolean... responseBodyAsAString) throws XExpression;

	/**
	 * Returns a bound partition list.
	 *
	 * @return Bound partition unique names
	 */
	Object getBoundPartitionNames();

	/**
	 * Issues an HTTP call to an external entity.
	 * <p>
     * If called from a logic element, it has the same timeout as defined for that element.
	 *
	 * @param url         Full URL to the HTTPS or HTTP endpoint.
	 * @param body        Body of the request.
	 * @param methodType  {@code "GET"} or {@code "POST"}
	 * @param contentType <ul>
	 *     				  <li>{@code "JSON"} - JSON response will be deserialized into a map</li>
	 * 					  <li>{@code "APPLICATION_XML"} (or {@code "XML"} (deprecated)) for contentType {@code "application/xml"}</li>
	 * 					  <li>{@code "TEXT_XML"} for contentType {@code "text/xml"}</li>
	 * 					  <li>Or any full content type, for instance {@code "text/plain"}</li>
	 * 					  </ul>
	 * 					  For XML content types, the response is
	 * 					  <a href="http://docs.groovy-lang.org/latest/html/api/groovy/util/slurpersupport/GPathResult.html">GPathResult</a> and
	 * 					  <a href="https://www.groovy-lang.org/processing-xml.html#_gpath">GPath</a> expressions.
	 *
	 * @param properties  Map of properties. The following are allowed:
	 *                    <ul>
	 *                    <li>basicAuthenticationUserName {@code String} value - User name for basic authentication.</li>
	 *                    <li>basicAuthenticationPassword {@code String} value - User password for basic authentication.</li>
	 *                    <li>sslTrustAllCertificates {@code boolean} value - Ignores invalid certificates. Use for testing HTTPS connections only. Not for production use.</li>
	 *                    <li>proxyAuthenticationUserName {@code String} value - User name for proxy (if any).</li>
	 *                    <li>proxyAuthenticationPassword {@code String} value - User password for proxy (if any).</li>
	 *                    <li>proxyUrl {@code String} value - Full URL to proxy.</li>
	 *                    <li>additionalHeaders {@code Map} value - Sub-map that will then add these headers verbatim to the request.</li>
	 *                    </ul>
	 * @return Map containing two elements: 'responseBody' and 'statusCode'.
	 * @throws XExpression If the HTTP call fails.
	 */
	Object httpCall(String url, String body, String methodType, String contentType, @SuppressWarnings("rawtypes") Map... properties) throws XExpression;

	/**
	 * Adds a new object to the collection of objects of the given type.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
     * This method cannot be used in a distributed calculation.
	 * <p>
	 * If any of the attributes is null, the record will not be updated.
	 * If you want a certain attribute to have a null value, do it explicitly, by putting e.g. the “NULL” marker value there or a similar one.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def entry = [
	 *         lookupTableName : "ExchangeRates",
	 *         key1 : "USD",
	 *         key2 : "EUR",
	 *         attribute1 : 0.879
	 * ]
	 * api.add("MLTV2", entry)
	 * </pre>
	 * </blockquote>
	 *
	 * @param typeCode Type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}.
	 * @param values   Key-value pairs representing fields of the object you want to add. Note that all significant fields have to be present.
	 * @return Added object or {@code null} if error occurred.
	 * @see #addOrUpdate(String, Map)
	 * @see #update(String, Map)
	 * @see #delete(String, Map)
	 */
	Object add(String typeCode, Map<String, Object> values);

	/**
	 * Adds or updates an object in the collection of objects of the given type.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
     * This method also works with the Post Step Logic Formula in workflows. This method cannot be used in a distributed calculation.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def entry = [
	 *         lookupTableName : "ExchangeRates",
	 *         key1 : "USD",
	 *         key2 : "EUR",
	 *         attribute1 : 0.879
	 * ]
	 * api.add("MLTV2", entry)
	 * </pre>
	 * </blockquote>
	 *
	 * @param typeCode Type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}.
	 * @param values   Key-value pairs representing fields of the object you want to add or update. Note that all significant fields have to be present.
	 * @return Object added or updated.
	 * @see #add(String, Map)
	 * @see #update(String, Map)
	 * @see #delete(String, Map)
	 */
	Object addOrUpdate(String typeCode, Map<String, Object> values);

	/**
	 * Updates an object in the collection of objects of the given type.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
     * This method cannot be used in a distributed calculation.
	 * <p>
	 * Example 1:
     * <blockquote>
	 * <pre>
	 * api.addOrUpdate("PX", [
	 * 		"name": "Scoring_MM",
	 * 		"sku": sku
	 * ])
	 * </pre>
	 * </blockquote>
     *
	 * Example 2:
	 * <blockquote>
	 * <pre>
	 * api.addOrUpdate("PGI", [
	 * 		"priceGridName" : "LifeCycleMonitor",
	 * 		"sku"           : currentItem.sku,
	 * 		"label"         : currentItem.label
	 * ])
	 * </pre>
	 * </blockquote>
	 *
	 * Example 3:
	 * <blockquote>
	 * <pre>
	 * def large = [
	 * 		"lookupTableId"   : pp.id,
	 * 		"lookupTableName" : pp.uniqueName,
	 * 		"name"            : "L",
	 * 		"attribute1"      : percentileHigh,
	 * 		"attribute2"      : "999999999"
	 * ]
     * api.addOrUpdate("MLTV", large)
	 * </pre>
	 * </blockquote>
     *
	 * @param typeCode Type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}.
	 * @param values   Key-value pairs representing fields of the object you want to update. Note that all significant fields have to be present.
	 * @return Updated object or {@code null} if error occured.
	 * @see #addOrUpdate(String, Map)
	 * @see #add(String, Map)
	 * @see #delete(String, Map)
	 */
	Object update(String typeCode, Map<String, Object> values);

	/**
	 * Deletes an object in the collection of objects of the given type.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
     * This method cannot be used in a distributed calculation.
	 * <p>
     * Example:
     * <blockquote>
	 * <pre>
	 * api.delete("PX", ["id": existingRecord, "name": "ActivePrice"])
	 * </pre>
	 * </blockquote>
	 *
	 * @param typeCode Type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}.
	 * @param values   Key-value pairs representing significant fields of the object you want to delete. Note that all significant fields have to be present; if not, the ID has to be present.
	 * @return Deleted object or {@code null} if error occured.
	 * @see #addOrUpdate(String, Map)
	 * @see #add(String, Map)
	 * @see #update(String, Map)
	 */
	Object delete(String typeCode, Map<String, Object> values);

	/**
	 * Sends a notification to Platform for further message routing.
	 *
	 * @param subject Subject of message.
	 * @param message Body of message.
	 */
	void sendPlatformNotification(String subject, String message);

	/**
	 * Sends an email. Only one recipient can be defined.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def message = "The cost is missing"
     * def filter = Filter.equals("name", "Costs")
	 *
	 * api.findLookupTableValues("EmailNotifications", filter).each {
	 *     api.sendEmail(it.value, "Missing costs", message)
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @param to      Email address the email is sent to.
	 * @param subject Email subject.
	 * @param message Email body.
	 */
	void sendEmail(String to, String subject, String message);

	/**
	 * Sends an email. Only one recipient can be defined.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
 	 *
	 * @param to		    Email address the email is sent to.
	 * @param toName		Name of the recipient. If not specified, toEmail is used.
	 * @param subject		Email subject.
	 * @param message		Email body.
	 * @param fromEmail 	Sender's email address.
	 * @param fromName		Name of the sender. If not specified, fromEmail is used.
	 * @param replyToEmail	Reply to email address.
	 * @param replyToName	Reply to user name. If not specified, replyToEmail is used.
	 * @param useCustomSMTP	True if sending through custom SMTP server should be used.
	 */
	void sendEmail(String to, String toName, String subject, String message, String fromEmail, String fromName, String replyToEmail, String replyToName, Boolean useCustomSMTP);

	/**
	 * Sends an event of the type "CUSTOM". Other event types are reserved internally.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CF and direct formula execution via JSON API).
	 * <p>
	 * Sending this type of events needs to be enabled on the partition by the system admin (event type "CUSTOM" in the Event Bitmask).
	 *
	 * @param object Object to pass along with the event.
	 */
	void customEvent(Object object);

	/**
	 * Sends an event of the type "CUSTOM". Other event types are reserved internally.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
	 * <p>
	 * Sending this type of events needs to be enabled on the partition by the system admin (event type "CUSTOM" in the Event Bitmask).
	 *
	 * @param object          Object to pass along with the event.
	 * @param customEventType Custom string to classify the custom event type (as the general event type is always CUSTOM).
	 */
	void customEvent(Object object, String customEventType);

	/**
	 * Sends an event of the type "CUSTOM". Other event types are reserved internally.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
	 * <p>
	 * Sending this type of events needs to be enabled on the partition by the system admin (event type "CUSTOM" in the Event Bitmask).
	 *
	 * @param object          Object to pass along with the event.
	 * @param customEventType Custom string to classify the custom event type (as the general event type is always CUSTOM).
	 * @param operation       One of these strings: UPDATE, DELETE, ADD, ITEMCHANGE.
	 */
	void customEvent(Object object, String customEventType, String operation);

	/**
	 * Sends an event of the type "CUSTOM". Other event types are reserved internally.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
	 * <p>
	 * Sending this type of events needs to be enabled on the partition by the system admin (event type "CUSTOM" in the Event Bitmask).
	 *
	 * @param object          Object to pass along with the event.
	 * @param customEventType Custom string to classify the custom event type (as the general event type is always CUSTOM). Can be null.
	 * @param operation       One of these strings: UPDATE, DELETE, ADD, ITEMCHANGE. Can be null.
	 * @param omitNullValues  Also includes keys that have a null value (false) or omits them (true). True is the default for other method signatures that do not explicitly specify that value.
	 */
	void customEvent(Object object, String customEventType, String operation, boolean omitNullValues);

	/**
	 * Updates the header information of a CalculableObject (e.g., a price grid, a price list, etc.).
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
	 *
	 * @param calcObjTypeCode Type code of the object to update.
	 * @param calcObjName     (Numeric) ID of the object to update.
	 * @return Calculable object header
	 */
	Object updateCalculableObjectHeader(String calcObjTypeCode, String calcObjName);

	/**
	 * Creates an {@link AttributedResult} object to be returned as a result of an output element.
	 * <p>
	 * Since {@code AttributedResult} is configurable, it is used when it is not sufficient to only return a
	 * plain value, but it is also required to adjust formatting, e.g. color, underline, etc.
	 * <p>
	 * The usage is simple: instead of returning the output value, create a new {@code AttributedResult} instance using
	 * {@code api.attributedResult(plain value)}, customize it by calling its various methods, and return it.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>{@code
	 * def cost = api.getElement("Margin_pct")
	 * return api.attributedResult(cost)
	 * .withBackgroundColor(cost < 0.30 ? "red" : "#0101DF")
	 * .withSuffix(cost < 0.30 ? "!!!" : null)
	 * .withTextColor(cost < 0.30 ? "white" : null)
	 * .withTextDecoration(cost < 0.30 ? "underline" : null)
	 * }</pre>
	 * </blockquote>
	 *
	 * @param result Calculation result.
	 * @return AttributedResult object
	 * @see AttributedResult AttributedResult - for details on formatting
	 */
	AttributedResult attributedResult(Object result);

	/**
	 * Wraps a result to make sure the original result object appears in the JSON response.
	 * This way also more complex results like lists and maps can be exported as a logic element result.
	 *
	 * @param result Result.
	 * @return Wrapped result object. Should only be used as a return statement of a logic element.
	 */
	WrappedResult wrap(Object result);

	/**
	 * Creates a new result matrix DTO object that can be further customized and populated with calculated data.
	 * The individual cells can be styled via the use of:
	 * <pre>
	 * -  linkCell(Object value, String targetPage, String targetPageState)
	 * -  imageCell(Object imageUrl)
	 * -  styledCell(Object value, String textColor, String bgColor, String weight, String alignment)
	 * -  styledCell(Object value, String textColor, String bgColor, String weight)
	 * -  styledCell(Object value, String textColor, String bgColor)
	 * -  setColumnFormat(String column, FieldFormatType fft)
	 * </pre>
	 * Valid FieldFormatTypes are:
	 * <pre>
	 * NUMERIC, NUMERIC_LONG, MONEY, PERCENT, TEXT, MONEY_EUR, MONEY_USD, MONEY_GBP, MONEY_JPY, MONEY_CHF, MONEY_PLN, DATETIME, DATE, INTEGER
	 * </pre>
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>{@code
	 * def getTrafficColor(value) {
	 *     if (value <= 0) {
	 *         return "red"
	 *     } else if (value > 0 && value < 0.1) {
	 *         return "yellow"
	 *     } else if (value >= 0.1) {
	 *         return "green"
	 *     }
	 * }
	 *
	 * // initialize new matrix with three columns
	 * def matrix = api.newMatrix("Item", "Quantity", "Price", "Pricegrid", "Pricelist", "Margin Status", "Image", "CustomerId")
	 *
	 * // add more columns dynamically
	 * matrix.addColumn("Margin %")
	 * matrix.addColumn("Url")
	 *
	 * // set column formats
	 * matrix.setColumnFormat("Item", FieldFormatType.TEXT)
	 * matrix.setColumnFormat("Quantity", FieldFormatType.NUMERIC_LONG)
	 * matrix.setColumnFormat("Price", FieldFormatType.MONEY_EUR)
	 * matrix.setColumnFormat("Margin %", FieldFormatType.PERCENT)
	 * matrix.setColumnFormat("Url", FieldFormatType.LINK)
	 *
	 * // allow users to filter values
	 * matrix.setEnableClientFilter(true)
	 *
	 * // add data row
	 * matrix.addRow([
	 *         "Item": matrix.styledCell("Red bold text", "#ff0000", "transparent", "bold"),
	 *         "Quantity": 1.23456,
	 *         "Price": 78.9012,
	 *         "Margin %": 0.3456,
	 *
	 *         "Url": "<a href=\"https://www.google.com/search?q=price+f(x)\">link</a>",
	 *
	 *         // optionally you can add link to a price grid of a given ID
	 *         "Pricegrid" : matrix.linkToPriceGrid("Open Pricegrid", 123, null),
	 *
	 *         // or to a price list of a given ID
	 *         "Pricelist" : matrix.linkToPriceList("Open Pricelist", 123, null),
	 *
	 *         // or a to custom page (result identical to the previous)
	 *         "CustomerId" : matrix.linkCell("Open Customer", "customersPage", "123456"),
	 *
	 *         // or add a library images, such as Traffic, BlackTraffic or Arrow
	 *         "Margin Status" : matrix.libraryImage("BlackTraffic", getTrafficColor(-0.2f)),
	 *
	 *         // or add a general image
	 *         "Image" : matrix.imageCell("images/grid/approve.png"),
	 * ])
	 *
	 *
	 * return matrix
	 * }</pre>
	 * </blockquote>
     *
	 * @return ResultMatrix object
	 * @see #newMatrix(String...)
	 */
	ResultMatrix newMatrix();

	/**
	 * Creates a new result matrix DTO object that can be further customized and populated with calculated data.
	 * <p>
	 * For details see {@link #newMatrix()}.
	 *
	 * @param columns Initial set of columns for the result matrix.
	 * @return ResultMatrix object
	 * @see #newMatrix()
	 */
	ResultMatrix newMatrix(String... columns);

	/**
	 * Creates a new result matrix DTO object that can be further customized and populated with calculated data.
	 * <p>
	 * For details see {@link #newMatrix()}.
	 *
	 * @see #newMatrix(String...)
	 * @param columns Initial set of columns for the result matrix.
	 *
	 * @return ResultMatrix object
	 * @see #newMatrix()
	 */
	ResultMatrix newMatrix(Collection<String> columns);

	/**
	 * Creates a new gauge object that can be further customized.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def gauge = api.newGauge()
	 * gauge.addSector(5, "#FF0000")
	 * gauge.addSector(25, "#FFFF00")
	 * gauge.addSector(null, "#00FF00")
	 * gauge.setValue(10)
	 * </pre>
	 * </blockquote>
	 * @return ResultGauge object
	 */
	ResultGauge newGauge();

	/**
	 * Returns a {@link DashboardApi} instance for a dashboard of the specified name.
	 * This method is typically used for configuration of
	 * <a href="https://pricefx.atlassian.net/wiki/search?text=Embedded%20Dashboard">Embedded Dashboards</a>.
	 * @param dashboardName name of the dashboard
	 * @return DashboardApi for specified dashboard
	 * @see DashboardApi
	 */
	Object dashboard(String dashboardName);

	/**
	 * <p>Constructs an event name that is unique to the current dashboard.
	 * </p>
	 * <p>This method is typically used for configuration of
	 * <a href="https://pricefx.atlassian.net/wiki/search?text=Embedded%20Dashboard">Embedded Dashboards</a>.
	 * </p>
	 * Events are global through browser tabs, so they are not limited to embedded dashboards.
	 * Every event has its source (e.g. a grid) and a payload – arbitrary data (e.g. values of grid columns),
	 * which are passed to the dashboard portlet. Elements which can trigger an embedded dashboard to receive new parameters are:
	 * <ul>
	 *   <li>ResultMatrix
	 *   <li>Open Embedded Dashboard on New Tab
	 *   <li>ScatterChart
	 *   <li>BarLineChart
	 *   <li>Data Tab of any Chart (except Waterfalls)
	 * </ul>
	 *
	 * @param eventName name of the event
	 * @return event name
	 */
	String dashboardWideEvent(String eventName);

	/**
	 * Creates a new dashboard controller object that can be further customized.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>{@code
	 * def lookupTable = api.findLookupTable("SalesOrg")
	 * def controller = api.newController()
	 * controller.addButton("1. STEP: Change Price Strategy", "pricingParametersPage", lookupTable.typedId)
	 * controller.addHTML("<h2>Meatball</h2>")
	 * }</pre>
	 * </blockquote>
	 *
	 * @return DashboardController object
	 */
	DashboardController newController();

	/**
	 * Walks a filter and all its sub-filters, visiting each filter in the tree.
	 * A pair of closures is used to visit each filter. Each closure can
	 * replace the filter that is visiting. If it does, a new tree will
	 * be created for every part of the tree that is affected, thus preserving
	 * the original tree.
	 * <p>
	 * {@code removeNulls} is consulted only in case of (sub)filters which take a list of sub-filters.
	 * <p>
	 * This method may be used in many cases. For instance, you
	 * can use it for mapping the current property set to another one, e.g. to convert
	 * a PB filter to a PA one. Similarly to this example:
	 * <blockquote>
	 * <pre>{@code
	 * def propertyMappings = ["sku": "ProductID"]
	 *
	 * api.walkFilter(pbFilter, null, { filter ->
	 *     if (filter != null && propertyMappings[filter.property] != null) {
	 *         filter.property = propertyMappings[filter.property]
	 *     }
	 *     filter
	 * }, false);
	 * }</pre>
	 * </blockquote>
	 *
	 * @param filter Filter object.
	 * @param visitBefore Before closure.
	 * @param visitAfter After closure.
	 * @param removeNulls Removes nulls.
	 * @return If any changes have been made, the new Filter; if not, the
	 * original Filter.
	 */
	Filter walkFilter(Filter filter, Closure<?> visitBefore, Closure<?> visitAfter, boolean removeNulls);

	/**
	 * @param dataLoadLabel Label of DMDataLoad as seen in the Data Load UI.
	 * @param dataLoadType  Load type of DMDataLoad as seen in the Data Load UI. Namely:
	 *                      <ul>
	 *                      <li>Customers</li>
	 *                      <li>Products</li>
	 *                      <li>Flush</li>
	 *                      <li>Calendar</li>
	 *                      <li>Refresh</li>
	 *                      <li>Calculation</li>
	 *                      <li>Truncate</li>
	 *                      <li>Simulation</li>
	 *                      <li>ModelCalculation</li>
	 *                      <li>InternalCopy</li>
	 *                      <li>IndexMaintenance</li>
	 *                      </ul>
	 * @param target        Target of DMDataLoad as seen in the Data Load UI.
	 * @return Data Load if and only if there is exactly one Data Load matching (fail early strategy).
	 */
	Object findDataLoad(String dataLoadLabel, String dataLoadType, String target);

	/**
	 * @param rebateRecordLoadLabel Label of the rebate record load as seen in the Rebate Record UI.
	 * @return rebateRecordLoad if and only if there is exactly one load matching (fail early strategy).
	 */
	Object findRebateRecordLoad(String rebateRecordLoadLabel);

	/**
	 * Returns the current partition name.
	 *
	 * @return Partition name
	 */
	String currentPartitionName();

	/**
	 * Provides a wrapper for the org.joda.time.DateTimeZone.forId(String tzString) method.
	 * For details see their documentation.
	 * <p>
	 * Example 1:
	 * <blockquote>
	 * <pre>
	 * def timezone1 = api.getTimeZone("Europe/Berlin")
	 * def timezone2 = api.getTimeZone("Europe/Prague")
	 * def timezone3 = api.getTimeZone("America/New_York")
	 * def timezone4 = api.getTimeZone("Australia/Brisbane")
	 * </pre>
	 * </blockquote>
	 *
	 * <p>
	 * Example 2:
	 * <blockquote>
	 * <pre>
	 * def startHour = 20
	 * def offset = api.getTimeZone("Europe/Prague").getOffset(new Date().getTime())
	 * def startHourUTC = startHour - offset.intValue()
	 * </pre>
	 * </blockquote>
	 *
	 * @param tzString Timezone string.
	 * @return DateTimeZone object
	 * @see <a href="https://en.wikipedia.org/wiki/List_of_tz_database_time_zones">https://en.wikipedia.org/wiki/List_of_tz_database_time_zones</a>
	 */
	DateTimeZone getTimeZone(String tzString);

	/**
	 * Generates a random UUID of the specified length.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * return uuid(15)   // returns e.g. "VcydxgltxrVZSTV"
	 * </pre>
	 * </blockquote>
	 *
	 * @param len Desired number of characters.
     * @return UUID string
     */
	String uuid(int len);
	/**
	 * Generates a random UUID of the specified length and radix.
	 * <p>
	 * Examples:
	 * <blockquote>
	 * <pre>
	 * return uuid(8, 2) // returns e.g. "01001010" (8 character ID, base=2)
	 *
	 * return uuid(8, 10)       // returns e.g. "47473046" (8 character ID, base=10)
	 *
	 * return uuid(8, 16)       // returns e.g. "098F4D35" (8 character ID, base=16)
	 * </pre>
	 * </blockquote>
	 *
	 * @param len Desired number of characters.
	 * @param radix Number of allowable values for each character (must be &lt;=62).
     * @return UUID string
	 */
	String uuid(int len, int radix);

	/**
	 * Generates a RFC4122, version 4 ID.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * return api.uuid()   // returns e.g. "92329D39-6F5C-4520-ABFC-AAB64544E172"
	 * </pre>
	 * </blockquote>
	 *
	 * @return UUID string
	 */
	String uuid();

	/**
	 * Constructs a new composite key object which can be used as a key in a map.
	 * Equality of two keys is defined by the pair-wise equality of the corresponding key values.
	 *
	 * @param keys Key values.
	 * @return MultiKey composed of the passed ordered key values.
	 */
	Object multiKey(Object... keys);

	/**
	 * Lists all roles available in the application.
	 *
	 * @return Map of [roleUniqueName : roleLabel] pairs.
     */
	Map<String, String> roles();

	/**
	 * Creates an entityRef value (usually db ID) for the given entity. The value can be stored in attributeXX fields
	 * and you can set its meta type to ENTITY REFERENCE if you want the UI to show the data as links.
	 *
     * @param po Object of a typedCode P, C, DCR, Q, CT, RBA.
	 * @return String that represents the linkable state of the entity.
     * @throws IllegalAccessException If {@code po} is of a different typeCode.
	 */
	String entityRef(PartitionedObject po) throws IllegalAccessException;

	/**
	 * Evaluates an arithmetic expression that is passed as a string.
	 * <p>
	 * Supported operations are:
	 * <ul>
	 *     <li>Addition {@code +}</li>
	 *     <li>Subtraction {@code -}</li>
	 *     <li>Multiplication {@code *}</li>
	 *     <li>Division {@code /}</li>
	 *     <li>Exponentiation {@code ^}</li>
	 *     <li>Square root {@code sqrt}</li>
	 * </ul>
	 *
	 * Operator precedence and associativity rules are respected.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * evalExpression("((10 - 3^2 + 1) * -sqrt(1 * 2 + 3 * 4)) / 4")
	 * </pre>
	 * </blockquote>
	 *
	 * @param mathExpression Mathematical expression.
	 * @return Result of the expression as a double.
	 */
	double evalExpression(String mathExpression);

	/**
	 * Stores a string value in the shared cache with the given key.
	 * <p>
	 * Note the "global" nature of the key.
	 *
	 * @param key Cache key.
	 * @param value Value to store.
	 * @see #getSharedCache(String)
	 * @see #removeSharedCache(String)
	 */
	void setSharedCache(String key, String value);

	/**
	 * Retrieves a value from the shared cache (and resets the TTL of the key).
	 *
	 * @param key Cache key.
	 *
	 * @return Value or {@code null} found in the shared cache.
	 * @see #getSharedCache(String)
	 */
	String getSharedCache(String key);

	/**
	 * Deletes a key from the shared cache.
	 *
	 * @param key Cache key.
	 * @see #setSharedCache(String, String)
	 */
	void removeSharedCache(String key);

	/**
	 * Provides a standard Base64 encoding of a string.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * api.base64Encode("Pricefx") // "UHJpY2VmeA=="
	 * </pre>
	 * </blockquote>
	 *
	 * @param string Plain text to be encoded.
	 * @return base64 Encoded string.
	 * @see #base64Decode(String)
	 * @see <a href="https://en.wikipedia.org/wiki/Base64">https://en.wikipedia.org/wiki/Base64</a>
	 */
	String base64Encode(String string);

	/**
	 * Decodes a base64 encoded string.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * api.base64Decode("UHJpY2VmeA==") // "Pricefx"
	 * </pre>
	 * </blockquote>
	 *
	 * @param base64String Base64 encoded text.
	 * @return Plain text string
	 * @see #base64Encode(String)
	 * @see <a href="https://en.wikipedia.org/wiki/Base64">https://en.wikipedia.org/wiki/Base64</a>
	 */
	String base64Decode(String base64String);

	/**
	 * Returns a list of product reference records showing which other objects, of the given type, reference these SKUs.
	 *
	 * @param typeCode Currently, only "Q" or "RBA" are supported.
	 * @param skus Set of SKU strings that should be searched for.
	 * @return List of reference records
	 */
	List<?> getProductReferences(String typeCode, Set<String> skus);

	/**
	 * Returns a list of SKUs based on the definition of the product group.
	 *
	 * @param pg Product group object.
	 * @return List of SKUs
	 */
	List<String> getSkusFromProductGroup(ProductGroup pg);

	/**
	 * Returns a list of Customer IDs based on the definition of the customer group.
	 *
	 * @param cg Customer group object.
	 * @return List of customer IDs
	 */
	List<String> getCustomerIdsFromCustomerGroup(CustomerGroup cg);

	/**
	 * Returns the language part of the current request's locale (e.g. "en").
	 *
	 * @return Language part string
	 */
	String getLocale();

	/**
	 * Returns the base URL component of the current instance.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * api.getBaseURL() // "https://www.pricefx.eu"
	 * </pre>
	 * </blockquote>
	 *
	 * @return Base URL
	 */
	String getBaseURL();

	CreationWorkflowDefinition newCreationWorkflow();

	CreationWorkflowStepDefinition newCreationWorkflowStep();

	/**
	 * Finds the next revision uniqueNames for the given Quotes/Contracts/RebateAgreements.
	 *
	 * @param typeCode Can be one of: 'Q', 'CT', 'RBA'.
	 * @param uniqueNames List of uniqueNames.
	 * @return Map of [uniqueName : List of nextRevUNs]
	 */
	Map<String, List<String>> findNextRevUNs(String typeCode, String... uniqueNames);

	/**
	 * Gives the current item context in a claim validation.
	 *
	 * @return Claim Context instance
	 */
	Object getClaimContext();

	/**
	 * Indicates whether the current calculation is recalculating the entire list or only parts of it.
	 * <p>
	 * Note: Currently only applicable for price grid calculations.
	 *
	 * @return True if the entire grid is calculated, false otherwise. {@code null} is returned if method is not applicable.
	 */
	Boolean isFullListCalculation();

	/**
	 * Creates a new background calculation job for the specified price grid.
	 * Either a full calculation (parameters == null) or a partial calculation (parameter entry "skusToRecalc" as a list of SKU strings).
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct logic execution via JSON API).
	 *
	 * @param priceGridId ID of the price grid to calculate.
	 * @param parameters Optional job parameters.
	 * @return ID of the created job tracker
	 */
	Long triggerPriceGridCalculation(Long priceGridId, Map<String, Object> parameters);

	/**
	 * Creates a new background calculation job for the specified CFS.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct logic execution via JSON API).
	 *
	 * @param cfsID ID of CalculatedFieldSet.
	 * @return ID of the created job tracker
	 */
	Long triggerCFSCalculation(Long cfsID);

	/**
	 * Provides a Map of the current execution context info variables.
	 * <p>
	 * The only currently supported variable is <code>commandName</code> (if any) that was used to invoke the current
	 * logic execution, e.g. "addproducts", "save", "price", "submit", "calculate".
	 *
	 * @return Context map
	 */
	Map currentContext();

	/**
	 * Resets specified fields on the currentItem object to null. This applies only to price grid items. Fields that cannot be found on the item are skipped.
	 *
	 * @param fields Field names that should be reset.
	 */
	void resetCurrentItemFields(String... fields);

	/**
	 * Evaluates the product (group) filter logic.
	 * @param logicName name of the logic
	 * @param logicParam parameter
	 * @return list of filters
	 */
	List<Filter> evalProductFilterLogic(String logicName, String logicParam);

	/**
	 * Evaluates the customer (group) filter logic.
	 * @param logicName name of the logic
	 * @param logicParam parameter
	 * @return list of filters
	 */
	List<Filter> evalCustomerFilterLogic(String logicName, String logicParam);


	/**
	 * Creates or updates a user input attribute on an LPG.
	 * <p>
	 * <b>Note</b>: Input attributes are saved to the extension columns (e.g. attribute12) - the same as the output elements.
	 * Keep this in mind, so that you do not accidentally overwrite any existing output element.
	 *
	 * @param fieldName Name of field to be used for storing values, e.g. attribute12.
	 * @param elementName Name of the column, e.g. color.
	 * @param label Label of the column, e.g. Color.
	 * @param priceGridId Price grid ID which can usually be obtained by using {@code api.currentItem("id")}.
	 * @return Created or updated attribute
	 */
	Object createOrUpdatePricegridInputAttribute(String fieldName, String elementName, String label, Long priceGridId);

	/**
	 * Deletes a price grid input attribute.
	 *
	 * @param id Id of the input attribute to delete.
	 * @return Deleted attribute object or null if no attrbiute was found.
	 */
	Object removePricegridInputAttribute(Long id);

	/**
	 * Creates or updates a user input attribute on a price list.
	 * <p>
	 * <b>Note</b>: Input attributes are saved to the extension columns (e.g. attribute12) - the same as the output elements.
	 * Keep this in mind, so that you do not accidentally overwrite any existing output element.
	 *
	 * @param fieldName Name of field to be used for storing values, e.g. attribute12.
	 * @param elementName Name of the column, e.g. color.
	 * @param label Label of the column, e.g. Color.
	 * @param pricelistId Price grid ID which can usually be obtained by using {@code api.currentItem("id")}.
	 * @return Created or updated attribute
	 */
	Object createOrUpdatePricelistInputAttribute(String fieldName, String elementName, String label, Long pricelistId);

	/**
	 * Deletes a price list input attribute.
	 *
	 * @param id Id of the input attribute to delete.
	 * @return Deleted attribute object or null if no attribute was found.
	 */
	Object removePricelistInputAttribute(Long id);

	/**
	 * Sets the PG calculation output that will be shown in the header section after the PG calculation.
	 *
	 * @param pricegridId Price grid ID which can usually be obtained by using {@code api.currentItem("id")}.
	 * @param name Name of the calculation output, e.g. sales.
	 * @param label	Label of the calculation output, e.g. Global Sales.
	 * @param value Calculated value, e.g. €2.3M.
	 * @param message Message containing additional information.
	 * @return Saved calculation output
	 */
	Object setPricegridCalculationOutput(Long pricegridId, String name, String label, String value, String message);

	/**
	 * Removes the PG calculation output.
	 *
	 * @param pricegridId Price grid ID which can usually be obtained by using {@code api.currentItem("id")}.
	 * @param name Name of the calculation output that will be removed.
	 */
	void removePricegridCalculationOutput(Long pricegridId, String name);

	/**
	 * Chart definition that will be used to construct and display a Highchart in the header section.
	 *
	 * @param definition Highchart definition as a map.
	 * @param pricegridId Price grid ID which can usually be obtained by using {@code api.currentItem("id")}.
	 * @return Saved chart definition
	 */
	Object setPricegridCalculationChart(final Map<String, ?> definition, Long pricegridId);

	/**
	 * Sets the PL calculation output that will be shown in the header section after the PL calculation.
	 *
	 * @param pricelistId Price list ID which can usually be obtained by using {@code api.currentItem("id")}.
	 * @param name Name of the calculation output, e.g. sales.
	 * @param label	Label of the calculation output, e.g. Global Sales.
	 * @param value Calculated value, e.g. €2.3M.
	 * @param message Message containing additional information.
	 * @return Saved calculation output
	 */
	Object setPricelistCalculationOutput(Long pricelistId, String name, String label, String value, String message);

	/**
	 * Removes the PL calculation output.
	 *
	 * @param pricelistId Price list ID which can usually be obtained by using {@code api.currentItem("id")}.
	 * @param name Name of the calculation output that will be removed.
	 */
	void removePricelistCalculationOutput(Long pricelistId, String name);

	/**
	 * Chart definition that will be used to construct and display a Highchart in the header section.
	 *
	 * @param definition Highchart definition as a map.
	 * @param pricelistId Price list ID which can usually be obtained by using {@code api.currentItem("id")}.
	 * @return Saved chart definition
	 */
	Object setPricelistCalculationChart(final Map<String, ?> definition, Long pricelistId);

	/**
	 * <p>Provides factory to build various types of inputs.
	 * </p>
	 *
	 * <blockquote>
	 *     <pre>

	 *         def factorInput = api.inputBuilderFactory()
	 *     			.createHiddenEntry("Factor")
	 *     			.buildContextParameter()
	 *
	 *			def nameInput = api.inputBuilderFactory()
	 *        		.createStringUserEntry("Name")
	 *        		.setLabel("Name")
	 *        		.buildContextParameter()
	 *
	 *     </pre>
	 * </blockquote>
	 *
	 * @return instance of {@link InputBuilderFactory} , factory to build inputs
	 */
	InputBuilderFactory inputBuilderFactory();

	/**
	 * Creates and returns a new ItemRecommendation object for Cross-sell and Upsell API.
	 *
	 * @param id Item ID, e.g. SKU, an entry with this ID must be present in master data.
	 * @param score Relevancy score, the higher the more relevant.
	 * @return Newly created ItemRecommendation object
	 */
	Object newItemRecommendation(String id, Double score);

	/**
	 * Finds documents (Quotes, Contracts, Rebate Agreements or Claims) attached to a particular entity, or a line item of a particular entity.
	 * @param typedId typedId of a entity
	 * @param lineId Id of a line item within the entity specified by typedId, or null if no line item.
	 * @param filters Optional list of filters (that will be evaluated using the AND operator).
	 *
	 * @return List of {@link AttachedDocument}
	 */
	List<AttachedDocument> findAttachedDocuments(String typedId, String lineId, Filter ... filters);

	/**
	 * Instantates a model evaluator, providing access to a previously set up model.
	 * @param modelName	The model's uniqueName
	 * @return	A new ModelEvaluator object
	 */
	Object modelEvaluator(String modelName);

	/**
	 * Writes a payload string (should be valid json to ensure compatibility with e.g. JSON API calls) into the specified table by means of primary key
	 * @param tableName Table to write into. Needs to exist. Keys need to match, otherwise call will fail
	 * @param keys Map with key/value pairs for all primary keys as defined in table
	 * @param jsonPayload payload as valid JSON string
	 */
	void writeKV(String tableName, Map<String, Object> keys, String jsonPayload);

	/**
	 * Writes a payload into the specified table by means of primary key. payload will be piped through JSON serialization
	 * @param tableName Table to write into. Needs to exist. Keys need to match, otherwise call will fail
	 * @param keys Map with key/value pairs for all primary keys as defined in table
	 * @param payload Payload object
	 */
	void writeKV(String tableName, Map<String, Object> keys, Object payload);

	/**
	 * Writes larger data sets into a table (comparable to loaddata)
	 * @param tableName Table to write into. Needs to exist. Keys need to match, otherwise call will fail
	 * @param header List and order of fields in data. Needs to include all primary keys AND a field called 'payload'
	 * @param data List of rows to be inserted with values being in order as defined in header param. Size of row entry always needs to match header defintion
	 */
	void writeKVBulk(String tableName, List<String> header, List<List> data);

	/**
	 * Retrieves a single value based on primary key match
	 * @param tableName Table to read from
	 * @param keys Map with key/value pairs for all primary keys as defined in table
	 * @return JSON string of payload
	 */
	String getKV(String tableName, Map<String, Object> keys);

	/**
	 * Retrieves multiple values from a table
	 * @param tableName Table to read from
	 * @param filters Filters in simple format. Key is the field name. Value is exact value match. If value starts with a '&lt;' or '&gt;' character a greater/lesser than filter is applied
	 * @param maxRows Numbers of rows to retrieve. Capped by a system default
	 * @return List of matching objects
	 */
	List<Map<String,Object>> fecthKV(String tableName, Map<String, Object> filters, int maxRows);
}
